# copyright 2003-2017 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""unit tests for module cubicweb.utils"""

import base64
import datetime
import decimal
import doctest
import re
try:
    from unittest2 import TestCase
except ImportError:  # Python3
    from unittest import TestCase

from six import PY2
from six.moves import range

from cubicweb import Binary, Unauthorized
from cubicweb.devtools.testlib import CubicWebTC
from cubicweb.utils import (make_uid, UStringIO, RepeatList, HTMLHead,
                            QueryCache)
from cubicweb.entity import Entity

try:
    from cubicweb.utils import CubicWebJsonEncoder, json
except ImportError:
    json = None


class MakeUidTC(TestCase):
    def test_1(self):
        self.assertNotEqual(make_uid('xyz'), make_uid('abcd'))
        self.assertNotEqual(make_uid('xyz'), make_uid('xyz'))

    def test_2(self):
        d = set()
        while len(d)<10000:
            uid = make_uid('xyz')
            if uid in d:
                self.fail(len(d))
            if re.match('\d', uid):
                self.fail('make_uid must not return something begining with '
                          'some numeric character, got %s' % uid)
            d.add(uid)


class TestQueryCache(TestCase):
    def test_querycache(self):
        c = QueryCache(ceiling=20)
        # write only
        for x in range(10):
            c[x] = x
        self.assertEqual(c._usage_report(),
                         {'transientcount': 0,
                          'itemcount': 10,
                          'permanentcount': 0})
        c = QueryCache(ceiling=10)
        # we should also get a warning
        for x in range(20):
            c[x] = x
        self.assertEqual(c._usage_report(),
                         {'transientcount': 0,
                          'itemcount': 10,
                          'permanentcount': 0})
        # write + reads
        c = QueryCache(ceiling=20)
        for n in range(4):
            for x in range(10):
                c[x] = x
                c[x]
        self.assertEqual(c._usage_report(),
                         {'transientcount': 10,
                          'itemcount': 10,
                          'permanentcount': 0})
        c = QueryCache(ceiling=20)
        for n in range(17):
            for x in range(10):
                c[x] = x
                c[x]
        self.assertEqual(c._usage_report(),
                         {'transientcount': 0,
                          'itemcount': 10,
                          'permanentcount': 10})
        c = QueryCache(ceiling=20)
        for n in range(17):
            for x in range(10):
                c[x] = x
                if n % 2:
                    c[x]
                if x % 2:
                    c[x]
        self.assertEqual(c._usage_report(),
                         {'transientcount': 5,
                          'itemcount': 10,
                          'permanentcount': 5})

class UStringIOTC(TestCase):
    def test_boolean_value(self):
        self.assertTrue(UStringIO())


class RepeatListTC(TestCase):

    def test_base(self):
        l = RepeatList(3, (1, 3))
        self.assertEqual(l[0], (1, 3))
        self.assertEqual(l[2], (1, 3))
        self.assertEqual(l[-1], (1, 3))
        self.assertEqual(len(l), 3)
        # XXX
        self.assertEqual(l[4], (1, 3))

        self.assertFalse(RepeatList(0, None))

    def test_slice(self):
        l = RepeatList(3, (1, 3))
        self.assertEqual(l[0:1], [(1, 3)])
        self.assertEqual(l[0:4], [(1, 3)]*3)
        self.assertEqual(l[:], [(1, 3)]*3)

    def test_iter(self):
        self.assertEqual(list(RepeatList(3, (1, 3))),
                          [(1, 3)]*3)

    def test_add(self):
        l = RepeatList(3, (1, 3))
        self.assertEqual(l + [(1, 4)], [(1, 3)]*3  + [(1, 4)])
        self.assertEqual([(1, 4)] + l, [(1, 4)] + [(1, 3)]*3)
        self.assertEqual(l + RepeatList(2, (2, 3)), [(1, 3)]*3 + [(2, 3)]*2)

        x = l + RepeatList(2, (1, 3))
        self.assertIsInstance(x, RepeatList)
        self.assertEqual(len(x), 5)
        self.assertEqual(x[0], (1, 3))

        x = l + [(1, 3)] * 2
        self.assertEqual(x, [(1, 3)] * 5)

    def test_eq(self):
        self.assertEqual(RepeatList(3, (1, 3)),
                          [(1, 3)]*3)

    def test_pop(self):
        l = RepeatList(3, (1, 3))
        l.pop(2)
        self.assertEqual(l, [(1, 3)]*2)


class JSONEncoderTC(TestCase):
    def setUp(self):
        if json is None:
            self.skipTest('json not available')

    def encode(self, value):
        return json.dumps(value, cls=CubicWebJsonEncoder)

    def test_encoding_dates(self):
        self.assertEqual(self.encode(datetime.datetime(2009, 9, 9, 20, 30)),
                          '"2009/09/09 20:30:00"')
        self.assertEqual(self.encode(datetime.date(2009, 9, 9)),
                          '"2009/09/09"')
        self.assertEqual(self.encode(datetime.time(20, 30)),
                          '"20:30:00"')

    def test_encoding_decimal(self):
        self.assertEqual(self.encode(decimal.Decimal('1.2')), '1.2')

    def test_encoding_bare_entity(self):
        e = Entity(None)
        e.cw_attr_cache['pouet'] = 'hop'
        e.eid = 2
        self.assertEqual(json.loads(self.encode(e)),
                          {'pouet': 'hop', 'eid': 2})

    def test_encoding_entity_in_list(self):
        e = Entity(None)
        e.cw_attr_cache['pouet'] = 'hop'
        e.eid = 2
        self.assertEqual(json.loads(self.encode([e])),
                          [{'pouet': 'hop', 'eid': 2}])

    def test_encoding_binary(self):
        for content in (b'he he', b'h\xe9 hxe9'):
            with self.subTest(content=content):
                encoded = self.encode(Binary(content))
                self.assertEqual(base64.b64decode(encoded), content)

    def test_encoding_unknown_stuff(self):
        self.assertEqual(self.encode(TestCase), 'null')


class HTMLHeadTC(CubicWebTC):

    def htmlhead(self, datadir_url):
        with self.admin_access.web_request() as req:
            base_url = u'http://test.fr/data/'
            req.datadir_url = base_url
            head = HTMLHead(req)
            return head

    def test_concat_urls(self):
        base_url = u'http://test.fr/data/'
        head = self.htmlhead(base_url)
        urls = [base_url + u'bob1.js',
                base_url + u'bob2.js',
                base_url + u'bob3.js']
        result = head.concat_urls(urls)
        expected = u'http://test.fr/data/??bob1.js,bob2.js,bob3.js'
        self.assertEqual(result, expected)

    def test_group_urls(self):
        base_url = u'http://test.fr/data/'
        head = self.htmlhead(base_url)
        urls_spec = [(base_url + u'bob0.js', None),
                     (base_url + u'bob1.js', None),
                     (u'http://ext.com/bob2.js', None),
                     (u'http://ext.com/bob3.js', None),
                     (base_url + u'bob4.css', 'all'),
                     (base_url + u'bob5.css', 'all'),
                     (base_url + u'bob6.css', 'print'),
                     (base_url + u'bob7.css', 'print'),
                     (base_url + u'bob8.css', ('all', u'[if IE 8]')),
                     (base_url + u'bob9.css', ('print', u'[if IE 8]'))
                     ]
        result = head.group_urls(urls_spec)
        expected = [(base_url + u'??bob0.js,bob1.js', None),
                    (u'http://ext.com/bob2.js', None),
                    (u'http://ext.com/bob3.js', None),
                    (base_url + u'??bob4.css,bob5.css', 'all'),
                    (base_url + u'??bob6.css,bob7.css', 'print'),
                    (base_url + u'bob8.css', ('all', u'[if IE 8]')),
                    (base_url + u'bob9.css', ('print', u'[if IE 8]'))
                    ]
        self.assertEqual(list(result), expected)

    def test_getvalue_with_concat(self):
        self.config.global_set_option('concat-resources', True)
        base_url = u'http://test.fr/data/'
        head = self.htmlhead(base_url)
        head.add_js(base_url + u'bob0.js')
        head.add_js(base_url + u'bob1.js')
        head.add_js(u'http://ext.com/bob2.js')
        head.add_js(u'http://ext.com/bob3.js')
        head.add_css(base_url + u'bob4.css')
        head.add_css(base_url + u'bob5.css')
        head.add_css(base_url + u'bob6.css', 'print')
        head.add_css(base_url + u'bob7.css', 'print')
        head.add_ie_css(base_url + u'bob8.css')
        head.add_ie_css(base_url + u'bob9.css', 'print', u'[if lt IE 7]')
        result = head.getvalue()
        expected = u"""<head>
<link rel="stylesheet" type="text/css" media="all" href="http://test.fr/data/??bob4.css,bob5.css"/>
<link rel="stylesheet" type="text/css" media="print" href="http://test.fr/data/??bob6.css,bob7.css"/>
<!--[if lt IE 8]>
<link rel="stylesheet" type="text/css" media="all" href="http://test.fr/data/bob8.css"/>
<!--[if lt IE 7]>
<link rel="stylesheet" type="text/css" media="print" href="http://test.fr/data/bob9.css"/>
<![endif]--> 
<script type="text/javascript" src="http://test.fr/data/??bob0.js,bob1.js"></script>
<script type="text/javascript" src="http://ext.com/bob2.js"></script>
<script type="text/javascript" src="http://ext.com/bob3.js"></script>
</head>
"""
        self.assertEqual(result, expected)

    def test_getvalue_without_concat(self):
        self.config.global_set_option('concat-resources', False)
        try:
            base_url = u'http://test.fr/data/'
            head = self.htmlhead(base_url)
            head.add_js(base_url + u'bob0.js')
            head.add_js(base_url + u'bob1.js')
            head.add_js(u'http://ext.com/bob2.js')
            head.add_js(u'http://ext.com/bob3.js')
            head.add_css(base_url + u'bob4.css')
            head.add_css(base_url + u'bob5.css')
            head.add_css(base_url + u'bob6.css', 'print')
            head.add_css(base_url + u'bob7.css', 'print')
            head.add_ie_css(base_url + u'bob8.css')
            head.add_ie_css(base_url + u'bob9.css', 'print', u'[if lt IE 7]')
            result = head.getvalue()
            expected = u"""<head>
<link rel="stylesheet" type="text/css" media="all" href="http://test.fr/data/bob4.css"/>
<link rel="stylesheet" type="text/css" media="all" href="http://test.fr/data/bob5.css"/>
<link rel="stylesheet" type="text/css" media="print" href="http://test.fr/data/bob6.css"/>
<link rel="stylesheet" type="text/css" media="print" href="http://test.fr/data/bob7.css"/>
<!--[if lt IE 8]>
<link rel="stylesheet" type="text/css" media="all" href="http://test.fr/data/bob8.css"/>
<!--[if lt IE 7]>
<link rel="stylesheet" type="text/css" media="print" href="http://test.fr/data/bob9.css"/>
<![endif]--> 
<script type="text/javascript" src="http://test.fr/data/bob0.js"></script>
<script type="text/javascript" src="http://test.fr/data/bob1.js"></script>
<script type="text/javascript" src="http://ext.com/bob2.js"></script>
<script type="text/javascript" src="http://ext.com/bob3.js"></script>
</head>
"""
            self.assertEqual(result, expected)
        finally:
            self.config.global_set_option('concat-resources', True)


def UnauthorizedTC(TestCase):

    def _test(self, func):
        self.assertEqual(func(Unauthorized()),
                         'You are not allowed to perform this operation')
        self.assertEqual(func(Unauthorized('a')),
                         'a')
        self.assertEqual(func(Unauthorized('a', 'b')),
                         'You are not allowed to perform a operation on b')
        self.assertEqual(func(Unauthorized('a', 'b', 'c')),
                         'a b c')

    def test_str(self):
        self._test(str)

    if PY2:
        def test_unicode(self):
            self._test(unicode)


def load_tests(loader, tests, ignore):
    import cubicweb.utils
    tests.addTests(doctest.DocTestSuite(cubicweb.utils))
    return tests


if __name__ == '__main__':
    import unittest
    unittest.main()



# copyright 2003-2011 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""WSGI request adapter for cubicweb

NOTE: each docstring tagged with ``COME FROM DJANGO`` means that
the code has been taken (or adapted) from Djanco source code :
  http://www.djangoproject.com/

"""



import tempfile

from io import BytesIO

from six.moves.urllib.parse import parse_qs

from cubicweb.multipart import (
    copy_file, parse_form_data, parse_options_header)
from cubicweb.web import RequestError
from cubicweb.web.request import CubicWebRequestBase
from cubicweb.wsgi import pformat, normalize_header


class CubicWebWsgiRequest(CubicWebRequestBase):
    """most of this code COMES FROM DJANGO
    """

    def __init__(self, environ, vreg):
        # self.vreg is used in get_posted_data, which is called before the
        # parent constructor.
        self.vreg = vreg

        self.environ = environ
        self.path = environ['PATH_INFO']
        self.method = environ['REQUEST_METHOD'].upper()

        # content_length "may be empty or absent"
        try:
            length = int(environ['CONTENT_LENGTH'])
        except (KeyError, ValueError):
            length = 0
        # wsgi.input is not seekable, so copy the request contents to a temporary file
        if length < 100000:
            self.content = BytesIO()
        else:
            self.content = tempfile.TemporaryFile()
        copy_file(environ['wsgi.input'], self.content, maxread=length)
        self.content.seek(0, 0)
        environ['wsgi.input'] = self.content

        headers_in = dict((normalize_header(k[5:]), v) for k, v in self.environ.items()
                          if k.startswith('HTTP_'))
        if 'CONTENT_TYPE' in environ:
            headers_in['Content-Type'] = environ['CONTENT_TYPE']

        post, files = self.get_posted_data()

        super(CubicWebWsgiRequest, self).__init__(vreg, post,
                                                  headers= headers_in)
        self.content = environ['wsgi.input']
        if files is not None:
            for key, part in files.iterallitems():
                self.form.setdefault(key, []).append((part.filename, part.file))
            # 3.16.4 backward compat
            for key in files.keys():
                if len(self.form[key]) == 1:
                    self.form[key] = self.form[key][0]

    def __repr__(self):
        # Since this is called as part of error handling, we need to be very
        # robust against potentially malformed input.
        form = pformat(self.form)
        meta = pformat(self.environ)
        return '<CubicWebWsgiRequest\FORM:%s,\nMETA:%s>' % \
            (form, meta)

    ## cubicweb request interface ################################################

    def http_method(self):
        """returns 'POST', 'GET', 'HEAD', etc."""
        return self.method

    def relative_path(self, includeparams=True):
        """return the normalized path of the request (ie at least relative
        to the instance's root, but some other normalization may be needed
        so that the returned path may be used to compare to generated urls

        :param includeparams:
           boolean indicating if GET form parameters should be kept in the path
        """
        path = self.environ['PATH_INFO']
        path = path[1:] # remove leading '/'
        if includeparams:
            qs = self.environ.get('QUERY_STRING')
            if qs:
                return '%s?%s' % (path, qs)

        return path

    ## wsgi request helpers ###################################################

    def get_posted_data(self):
        # The WSGI spec says 'QUERY_STRING' may be absent.
        post = parse_qs(self.environ.get('QUERY_STRING', ''))
        files = None
        if self.method == 'POST':
            content_type = self.environ.get('CONTENT_TYPE')
            if not content_type:
                raise RequestError("Missing Content-Type")
            content_type, options = parse_options_header(content_type)
            if content_type in (
                    'multipart/form-data',
                    'application/x-www-form-urlencoded',
                    'application/x-url-encoded'):
                forms, files = parse_form_data(
                    self.environ, strict=True,
                    mem_limit=self.vreg.config['max-post-length'])
                post.update(forms.dict)
        self.content.seek(0, 0)
        return post, files

    def setup_params(self, params):
        # This is a copy of CubicWebRequestBase.setup_params, but without
        # converting unicode strings because it is partially done by
        # get_posted_data
        self.form = {}
        if params is None:
            return
        encoding = self.encoding
        for param, val in params.items():
            if isinstance(val, (tuple, list)):
                if len(val) == 1:
                    val = val[0]
            if param in self.no_script_form_params and val:
                val = self.no_script_form_param(param, val)
            if param == '_cwmsgid':
                self.set_message_id(val)
            else:
                self.form[param] = val



# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""cubicweb.migration unit tests"""

from os.path import dirname, join
from logilab.common.testlib import TestCase, unittest_main

from cubicweb import devtools
from cubicweb.cwconfig import CubicWebConfiguration
from cubicweb.migration import filter_scripts, version_strictly_lower


class Schema(dict):
    def has_entity(self, e_type):
        return e_type in self

SMIGRDIR = join(dirname(__file__), 'data', 'server_migration') + '/'
TMIGRDIR = join(dirname(__file__), 'data', 'migration') + '/'


class MigrTestConfig(devtools.TestServerConfiguration):
    verbosity = 0

    def migration_scripts_dir(cls):
        return SMIGRDIR

    def cube_migration_scripts_dir(cls, cube):
        return TMIGRDIR

class MigrationToolsTC(TestCase):
    def setUp(self):
        self.config = MigrTestConfig('data', __file__)
        from yams.schema import Schema
        self.config.load_schema = lambda expand_cubes=False: Schema('test')
        self.config.__class__.cubicweb_appobject_path = frozenset()
        self.config.__class__.cube_appobject_path = frozenset()

    def test_filter_scripts_base(self):
        self.assertListEqual(filter_scripts(self.config, SMIGRDIR, (2,3,0), (2,4,0)),
                              [])
        self.assertListEqual(filter_scripts(self.config, SMIGRDIR, (2,4,0), (2,5,0)),
                              [((2, 5, 0), SMIGRDIR+'2.5.0_Any.sql')])
        self.assertListEqual(filter_scripts(self.config, SMIGRDIR, (2,5,0), (2,6,0)),
                              [((2, 6, 0), SMIGRDIR+'2.6.0_Any.sql')])
        self.assertListEqual(filter_scripts(self.config, SMIGRDIR, (2,4,0), (2,6,0)),
                              [((2, 5, 0), SMIGRDIR+'2.5.0_Any.sql'),
                               ((2, 6, 0), SMIGRDIR+'2.6.0_Any.sql')])
        self.assertListEqual(filter_scripts(self.config, SMIGRDIR, (2,5,0), (2,5,1)),
                              [])
        self.assertListEqual(filter_scripts(self.config, SMIGRDIR, (2,5,0), (2,10,2)),
                              [((2, 6, 0), SMIGRDIR+'2.6.0_Any.sql'),
                               ((2, 10, 2), SMIGRDIR+'2.10.2_Any.sql')])
        self.assertListEqual(filter_scripts(self.config, SMIGRDIR, (2,5,1), (2,6,0)),
                              [((2, 6, 0), SMIGRDIR+'2.6.0_Any.sql')])

        self.assertListEqual(filter_scripts(self.config, TMIGRDIR, (0,0,2), (0,0,3)),
                              [((0, 0, 3), TMIGRDIR+'0.0.3_Any.py')])
        self.assertListEqual(filter_scripts(self.config, TMIGRDIR, (0,0,2), (0,0,4)),
                              [((0, 0, 3), TMIGRDIR+'0.0.3_Any.py'),
                               ((0, 0, 4), TMIGRDIR+'0.0.4_Any.py')])

    def test_filter_scripts_for_mode(self):
        config = CubicWebConfiguration('data', __file__)
        config.verbosity = 0
        config = self.config
        config.__class__.name = 'repository'
        self.assertListEqual(filter_scripts(config, TMIGRDIR, (0,0,4), (0,1,0)),
                              [((0, 1 ,0), TMIGRDIR+'0.1.0_Any.py'),
                               ((0, 1 ,0), TMIGRDIR+'0.1.0_common.py'),
                               ((0, 1 ,0), TMIGRDIR+'0.1.0_repository.py')])
        config.__class__.name = 'all-in-one'
        self.assertListEqual(filter_scripts(config, TMIGRDIR, (0,0,4), (0,1,0)),
                              [((0, 1 ,0), TMIGRDIR+'0.1.0_Any.py'),
                               ((0, 1 ,0), TMIGRDIR+'0.1.0_common.py'),
                               ((0, 1 ,0), TMIGRDIR+'0.1.0_repository.py')])
        config.__class__.name = 'repository'

    def test_version_strictly_lower(self):
        self.assertTrue(version_strictly_lower(None, '1.0.0'))
        self.assertFalse(version_strictly_lower('1.0.0', None))


from cubicweb.devtools import ApptestConfiguration, get_test_db_handler

class BaseCreationTC(TestCase):

    def test_db_creation(self):
        """make sure database can be created"""
        config = ApptestConfiguration('data', __file__)
        source = config.system_source_config
        self.assertEqual(source['db-driver'], 'sqlite')
        handler = get_test_db_handler(config)
        handler.init_test_database()
        handler.build_db_cache()
        repo, cnx = handler.get_repo_and_cnx()
        with cnx:
            self.assertEqual(cnx.execute('Any SN WHERE X is CWUser, X login "admin", X in_state S, S name SN').rows,
                             [['activated']])
        repo.shutdown()

if __name__ == '__main__':
    unittest_main()



# copyright 2003-2017 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""schema hooks:

- synchronize the living schema object with the persistent schema
- perform physical update on the source when necessary

checking for schema consistency is done in hooks.py
"""

import json
from copy import copy

from yams.schema import BASE_TYPES, BadSchemaDefinition, RelationDefinitionSchema
from yams.constraints import UniqueConstraint
from yams import buildobjs as ybo, convert_default_value

from logilab.common.decorators import clear_cache

from cubicweb import _
from cubicweb import validation_error
from cubicweb.predicates import is_instance
from cubicweb.schema import (SCHEMA_TYPES, META_RTYPES, VIRTUAL_RTYPES,
                             CONSTRAINTS, UNIQUE_CONSTRAINTS, ETYPE_NAME_MAP)
from cubicweb.server import hook, schemaserial as ss, schema2sql as y2sql
from cubicweb.server.sqlutils import SQL_PREFIX
from cubicweb.hooks.synccomputed import RecomputeAttributeOperation

# core entity and relation types which can't be removed
CORE_TYPES = BASE_TYPES | SCHEMA_TYPES | META_RTYPES | set(
    ('CWUser', 'CWGroup', 'login', 'upassword', 'name', 'in_group'))


def get_constraints(cnx, entity):
    constraints = []
    for cstreid in cnx.transaction_data.get(entity.eid, ()):
        cstrent = cnx.entity_from_eid(cstreid)
        cstr = CONSTRAINTS[cstrent.type].deserialize(cstrent.value)
        cstr.eid = cstreid
        constraints.append(cstr)
    return constraints


def group_mapping(cw):
    try:
        return cw.transaction_data['groupmap']
    except KeyError:
        cw.transaction_data['groupmap'] = gmap = ss.group_mapping(cw)
        return gmap


def add_inline_relation_column(cnx, etype, rtype):
    """add necessary column and index for an inlined relation"""
    attrkey = '%s.%s' % (etype, rtype)
    createdattrs = cnx.transaction_data.setdefault('createdattrs', set())
    if attrkey in createdattrs:
        return
    createdattrs.add(attrkey)
    table = SQL_PREFIX + etype
    column = SQL_PREFIX + rtype
    try:
        cnx.system_sql(str('ALTER TABLE %s ADD %s integer REFERENCES entities (eid)'
                           % (table, column)),
                       rollback_on_failure=False)
        cnx.info('added column %s to table %s', column, table)
    except Exception:
        # silent exception here, if this error has not been raised because the
        # column already exists, index creation will fail anyway
        cnx.exception('error while adding column %s to table %s',
                      table, column)
    # create index before alter table which may expectingly fail during test
    # (sqlite) while index creation should never fail (test for index existence
    # is done by the dbhelper)
    cnx.repo.system_source.create_index(cnx, table, column)
    cnx.info('added index on %s(%s)', table, column)


def insert_rdef_on_subclasses(cnx, eschema, rschema, rdefdef, props):
    # XXX 'infered': True/False, not clear actually
    props.update({'constraints': rdefdef.constraints,
                  'description': rdefdef.description,
                  'cardinality': rdefdef.cardinality,
                  'permissions': rdefdef.get_permissions(),
                  'order': rdefdef.order,
                  'infered': False, 'eid': None
                  })
    cstrtypemap = ss.cstrtype_mapping(cnx)
    groupmap = group_mapping(cnx)
    object = rschema.schema.eschema(rdefdef.object)
    for specialization in eschema.specialized_by(False):
        if (specialization, rdefdef.object) in rschema.rdefs:
            continue
        sperdef = RelationDefinitionSchema(specialization, rschema,
                                           object, None, values=props)
        ss.execschemarql(cnx.execute, sperdef,
                         ss.rdef2rql(sperdef, cstrtypemap, groupmap))


def check_valid_changes(cnx, entity, ro_attrs=('name', 'final')):
    errors = {}
    # don't use getattr(entity, attr), we would get the modified value if any
    for attr in entity.cw_edited:
        if attr in ro_attrs:
            origval, newval = entity.cw_edited.oldnewvalue(attr)
            if newval != origval:
                errors[attr] = _("can't change this attribute")
    if errors:
        raise validation_error(entity, errors)


class _MockEntity(object):  # XXX use a named tuple with python 2.6
    def __init__(self, eid):
        self.eid = eid


class SyncSchemaHook(hook.Hook):
    """abstract class for schema synchronization hooks (in the `syncschema`
    category)
    """
    __abstract__ = True
    category = 'syncschema'


# operations for low-level database alteration  ################################

class DropTable(hook.Operation):
    """actually remove a database from the instance's schema"""
    table = None  # make pylint happy

    def precommit_event(self):
        dropped = self.cnx.transaction_data.setdefault('droppedtables', set())
        if self.table in dropped:
            return  # already processed
        dropped.add(self.table)
        self.cnx.system_sql('DROP TABLE %s' % self.table)
        self.info('dropped table %s', self.table)

    # XXX revertprecommit_event


class DropRelationTable(DropTable):
    def __init__(self, cnx, rtype):
        super(DropRelationTable, self).__init__(
            cnx, table='%s_relation' % rtype)
        cnx.transaction_data.setdefault('pendingrtypes', set()).add(rtype)


class DropColumn(hook.DataOperationMixIn, hook.Operation):
    """actually remove the attribut's column from entity table in the system
    database
    """
    def precommit_event(self):
        cnx = self.cnx
        for etype, attr in self.get_data():
            table = SQL_PREFIX + etype
            column = SQL_PREFIX + attr
            source = cnx.repo.system_source
            # drop index if any
            source.drop_index(cnx, table, column)
            if source.dbhelper.alter_column_support:
                cnx.system_sql('ALTER TABLE %s DROP COLUMN %s' % (table, column),
                               rollback_on_failure=False)
                self.info('dropped column %s from table %s', column, table)
            else:
                # not supported by sqlite for instance
                self.error('dropping column not supported by the backend, handle '
                           'it yourself (%s.%s)', table, column)

    # XXX revertprecommit_event


# base operations for in-memory schema synchronization  ########################

class MemSchemaNotifyChanges(hook.SingleLastOperation):
    """the update schema operation:

    special operation which should be called once and after all other schema
    operations. It will trigger internal structures rebuilding to consider
    schema changes.
    """

    def __init__(self, cnx):
        hook.SingleLastOperation.__init__(self, cnx)

    def precommit_event(self):
        for eschema in self.cnx.repo.schema.entities():
            if not eschema.final:
                clear_cache(eschema, 'ordered_relations')
                # clear additional cached properties
                try:
                    # is_composite use composite_rdef_roles, hence one try except should be enough
                    del eschema.composite_rdef_roles
                    del eschema.is_composite
                except AttributeError:
                    pass

    def postcommit_event(self):
        repo = self.cnx.repo
        # commit event should not raise error, while set_schema has chances to
        # do so because it triggers full vreg reloading
        try:
            repo.schema.rebuild_infered_relations()
            # trigger vreg reload
            repo.set_schema(repo.schema)
        except Exception:
            self.critical('error while setting schema', exc_info=True)

    def rollback_event(self):
        self.precommit_event()


class MemSchemaOperation(hook.Operation):
    """base class for schema operations"""
    def __init__(self, cnx, **kwargs):
        hook.Operation.__init__(self, cnx, **kwargs)
        # every schema operation is triggering a schema update
        MemSchemaNotifyChanges(cnx)


# operations for high-level source database alteration  ########################

class CWETypeAddOp(MemSchemaOperation):
    """after adding a CWEType entity:
    * add it to the instance's schema
    * create the necessary table
    * set creation_date and modification_date by creating the necessary
      CWAttribute entities
    * add <meta rtype> relation by creating the necessary CWRelation entity
    """
    entity = None  # make pylint happy

    def precommit_event(self):
        cnx = self.cnx
        entity = self.entity
        schema = cnx.vreg.schema
        etype = ybo.EntityType(eid=entity.eid, name=entity.name,
                               description=entity.description)
        eschema = schema.add_entity_type(etype)
        # create the necessary table
        for sql in y2sql.eschema2sql(cnx.repo.system_source.dbhelper,
                                     eschema, prefix=SQL_PREFIX):
            cnx.system_sql(sql)
        # add meta relations
        gmap = group_mapping(cnx)
        cmap = ss.cstrtype_mapping(cnx)
        for rtype in (META_RTYPES - VIRTUAL_RTYPES):
            try:
                rschema = schema[rtype]
            except KeyError:
                self.critical('rtype %s was not handled at cwetype creation time', rtype)
                continue
            if not rschema.rdefs:
                self.warning('rtype %s has no relation definition yet', rtype)
                continue
            sampletype = rschema.subjects()[0]
            desttype = rschema.objects()[0]
            try:
                rdef = copy(rschema.rdef(sampletype, desttype))
            except KeyError:
                # this combo does not exist because this is not a universal META_RTYPE
                continue
            rdef.subject = _MockEntity(eid=entity.eid)
            mock = _MockEntity(eid=None)
            ss.execschemarql(cnx.execute, mock, ss.rdef2rql(rdef, cmap, gmap))

    def revertprecommit_event(self):
        # revert changes on in memory schema
        self.cnx.vreg.schema.del_entity_type(self.entity.name)
        # revert changes on database
        self.cnx.system_sql('DROP TABLE %s%s' % (SQL_PREFIX, self.entity.name))


class CWETypeRenameOp(MemSchemaOperation):
    """this operation updates physical storage accordingly"""

    oldname = newname = None  # make pylint happy

    def rename(self, oldname, newname):
        cnx = self.cnx
        source = cnx.repo.system_source
        dbhelper = source.dbhelper
        # we need sql to operate physical changes on the system database
        sqlexec = cnx.system_sql
        cnx.vreg.schema.rename_entity_type(oldname, newname)
        old_table = SQL_PREFIX + oldname
        new_table = SQL_PREFIX + newname
        eschema = cnx.vreg.schema.eschema(newname)
        # drop old indexes before the renaming
        for rschema in eschema.subject_relations():
            if rschema.inlined or (rschema.final and eschema.rdef(rschema.type).indexed):
                source.drop_index(cnx, old_table, SQL_PREFIX + rschema.type)
            if rschema.final and any(isinstance(cstr, UniqueConstraint)
                                     for cstr in eschema.rdef(rschema.type).constraints):
                source.drop_index(cnx, old_table, SQL_PREFIX + rschema.type, unique=True)
        sql = dbhelper.sql_rename_table(old_table, new_table)
        sqlexec(sql)
        self.info('renamed table %s to %s', oldname, newname)
        sqlexec('UPDATE entities SET type=%(newname)s WHERE type=%(oldname)s',
                {'newname': newname, 'oldname': oldname})
        for eid, etype in cnx.repo._type_cache.items():
            if etype == oldname:
                cnx.repo._type_cache[eid] = newname
        # recreate the indexes
        for rschema in eschema.subject_relations():
            if rschema.inlined or (rschema.final and eschema.rdef(rschema.type).indexed):
                source.create_index(cnx, new_table, SQL_PREFIX + rschema.type)
            if rschema.final and any(isinstance(cstr, UniqueConstraint)
                                     for cstr in eschema.rdef(rschema.type).constraints):
                source.create_index(cnx, new_table, SQL_PREFIX + rschema.type, unique=True)
        for attrs in eschema._unique_together or ():
            columns = ['%s%s' % (SQL_PREFIX, attr) for attr in attrs]
            old_index_name = y2sql.unique_index_name(oldname, columns)
            for sql in dbhelper.sqls_drop_multicol_unique_index(
                    new_table, columns, old_index_name):
                sqlexec(sql)
            new_index_name = y2sql.unique_index_name(newname, columns)
            for sql in dbhelper.sqls_create_multicol_unique_index(
                    new_table, columns, new_index_name):
                sqlexec(sql)
        # XXX transaction records

    def precommit_event(self):
        self.rename(self.oldname, self.newname)

    def revertprecommit_event(self):
        self.rename(self.newname, self.oldname)


class CWRTypeUpdateOp(MemSchemaOperation):
    """actually update some properties of a relation definition"""

    rschema = entity = values = None  # make pylint happy
    oldvalues = None

    def precommit_event(self):
        rschema = self.rschema
        if rschema.final:
            return  # watched changes to final relation type are unexpected
        cnx = self.cnx
        if 'fulltext_container' in self.values:
            op = UpdateFTIndexOp.get_instance(cnx)
            for subjtype, objtype in rschema.rdefs:
                if self.values['fulltext_container'] == 'subject':
                    op.add_data(subjtype)
                    op.add_data(objtype)
                else:
                    op.add_data(objtype)
                    op.add_data(subjtype)
        # update the in-memory schema first
        self.oldvalues = dict((attr, getattr(rschema, attr)) for attr in self.values)
        self.rschema.__dict__.update(self.values)
        # then make necessary changes to the system source database
        if 'inlined' not in self.values:
            return  # nothing to do
        inlined = self.values['inlined']
        # check in-lining is possible when inlined
        if inlined:
            self.entity.check_inlined_allowed()
        # inlined changed, make necessary physical changes!
        sqlexec = self.cnx.system_sql
        rtype = rschema.type
        eidcolumn = SQL_PREFIX + 'eid'
        if not inlined:
            # need to create the relation if it has not been already done by
            # another event of the same transaction
            if rschema.type not in cnx.transaction_data.get('createdtables', ()):
                # create the necessary table
                for sql in y2sql.rschema2sql(rschema):
                    sqlexec(sql)
                cnx.transaction_data.setdefault('createdtables', []).append(
                    rschema.type)
            # copy existant data
            column = SQL_PREFIX + rtype
            for etype in rschema.subjects():
                table = SQL_PREFIX + str(etype)
                sqlexec('INSERT INTO %s_relation SELECT %s, %s FROM %s WHERE NOT %s IS NULL'
                        % (rtype, eidcolumn, column, table, column))
            # drop existant columns
            for etype in rschema.subjects():
                DropColumn.get_instance(cnx).add_data((str(etype), rtype))
        else:
            for etype in rschema.subjects():
                try:
                    add_inline_relation_column(cnx, str(etype), rtype)
                except Exception as ex:
                    # the column probably already exists. this occurs when the
                    # entity's type has just been added or if the column has not
                    # been previously dropped (eg sqlite)
                    self.error('error while altering table %s: %s', etype, ex)
                # copy existant data.
                # XXX don't use, it's not supported by sqlite (at least at when i tried it)
                #sqlexec('UPDATE %(etype)s SET %(rtype)s=eid_to '
                #        'FROM %(rtype)s_relation '
                #        'WHERE %(etype)s.eid=%(rtype)s_relation.eid_from'
                #        % locals())
                table = SQL_PREFIX + str(etype)
                cursor = sqlexec('SELECT eid_from, eid_to FROM %(table)s, '
                                 '%(rtype)s_relation WHERE %(table)s.%(eidcolumn)s='
                                 '%(rtype)s_relation.eid_from' % locals())
                args = [{'val': eid_to, 'x': eid} for eid, eid_to in cursor.fetchall()]
                if args:
                    column = SQL_PREFIX + rtype
                    cursor.executemany('UPDATE %s SET %s=%%(val)s WHERE %s=%%(x)s'
                                       % (table, column, eidcolumn), args)
                # drop existant table
                DropRelationTable(cnx, rtype)

    def revertprecommit_event(self):
        # revert changes on in memory schema
        self.rschema.__dict__.update(self.oldvalues)
        # XXX revert changes on database


class CWComputedRTypeUpdateOp(MemSchemaOperation):
    """actually update some properties of a computed relation definition"""
    rschema = entity = rule = None  # make pylint happy
    old_rule = None

    def precommit_event(self):
        # update the in-memory schema first
        self.old_rule = self.rschema.rule
        self.rschema.rule = self.rule

    def revertprecommit_event(self):
        # revert changes on in memory schema
        self.rschema.rule = self.old_rule


class CWAttributeAddOp(MemSchemaOperation):
    """an attribute relation (CWAttribute) has been added:
    * add the necessary column
    * set default on this column if any and possible
    * register an operation to add the relation definition to the
      instance's schema on commit

    constraints are handled by specific hooks
    """
    entity = None  # make pylint happy

    def init_rdef(self, **kwargs):
        entity = self.entity
        fromentity = entity.stype
        rdefdef = self.rdefdef = ybo.RelationDefinition(
            str(fromentity.name), entity.rtype.name, str(entity.otype.name),
            description=entity.description, cardinality=entity.cardinality,
            constraints=get_constraints(self.cnx, entity),
            order=entity.ordernum, eid=entity.eid, **kwargs)
        try:
            self.cnx.vreg.schema.add_relation_def(rdefdef)
        except BadSchemaDefinition:
            # rdef has been infered then explicitly added (current consensus is
            # not clear at all versus infered relation handling (and much
            # probably buggy)
            rdef = self.cnx.vreg.schema.rschema(rdefdef.name).rdefs[rdefdef.subject, rdefdef.object]
            assert rdef.infered
        else:
            rdef = self.cnx.vreg.schema.rschema(rdefdef.name).rdefs[rdefdef.subject, rdefdef.object]

        self.cnx.execute('SET X ordernum Y+1 '
                         'WHERE X from_entity SE, SE eid %(se)s, X ordernum Y, '
                         'X ordernum >= %(order)s, NOT X eid %(x)s',
                         {'x': entity.eid, 'se': fromentity.eid,
                          'order': entity.ordernum or 0})
        return rdefdef, rdef

    def precommit_event(self):
        cnx = self.cnx
        entity = self.entity
        # entity.defaultval is a Binary or None, but we need a correctly typed
        # value
        default = entity.defaultval
        if default is not None:
            default = default.unzpickle()
        props = {'default': default,
                 'indexed': entity.indexed,
                 'fulltextindexed': entity.fulltextindexed,
                 'internationalizable': entity.internationalizable}
        if entity.extra_props:
            props.update(json.loads(entity.extra_props.getvalue().decode('ascii')))
        # entity.formula may not exist yet if we're migrating to 3.20
        if hasattr(entity, 'formula'):
            props['formula'] = entity.formula
        # update the in-memory schema first
        rdefdef, rdef = self.init_rdef(**props)
        # then make necessary changes to the system source database
        syssource = cnx.repo.system_source
        attrtype = y2sql.type_from_rdef(syssource.dbhelper, rdef)
        # added some str() wrapping query since some backend (eg psycopg) don't
        # allow unicode queries
        table = SQL_PREFIX + rdefdef.subject
        column = SQL_PREFIX + rdefdef.name
        try:
            cnx.system_sql(str('ALTER TABLE %s ADD %s %s'
                               % (table, column, attrtype)),
                           rollback_on_failure=False)
            self.info('added column %s to table %s', column, table)
        except Exception as ex:
            # the column probably already exists. this occurs when
            # the entity's type has just been added or if the column
            # has not been previously dropped
            self.error('error while altering table %s: %s', table, ex)
        if entity.indexed:
            try:
                syssource.create_index(cnx, table, column, unique=False)
            except Exception as ex:
                self.error('error while creating index for %s.%s: %s',
                           table, column, ex)
        # final relations are not infered, propagate
        schema = cnx.vreg.schema
        try:
            eschema = schema.eschema(rdefdef.subject)
        except KeyError:
            return  # entity type currently being added
        # propagate attribute to children classes
        rschema = schema.rschema(rdefdef.name)
        # if relation type has been inserted in the same transaction, its final
        # attribute is still set to False, so we've to ensure it's False
        rschema.final = True
        insert_rdef_on_subclasses(cnx, eschema, rschema, rdefdef, props)
        # update existing entities with the default value of newly added attribute
        if default is not None:
            default = convert_default_value(self.rdefdef, default)
            cnx.system_sql('UPDATE %s SET %s=%%(default)s' % (table, column),
                           {'default': default})
        # if attribute is computed, compute it
        if getattr(entity, 'formula', None):
            # add rtype attribute for RelationDefinitionSchema api compat, this
            # is what RecomputeAttributeOperation expect
            rdefdef.rtype = rdefdef.name
            RecomputeAttributeOperation.get_instance(cnx).add_data(rdefdef)

    def revertprecommit_event(self):
        # revert changes on in memory schema
        if getattr(self, 'rdefdef', None) is None:
            return
        self.cnx.vreg.schema.del_relation_def(
            self.rdefdef.subject, self.rdefdef.name, self.rdefdef.object)
        # XXX revert changes on database


class CWRelationAddOp(CWAttributeAddOp):
    """an actual relation has been added:

    * add the relation definition to the instance's schema

    * if this is an inlined relation, add the necessary column else if it's the
      first instance of this relation type, add the necessary table and set
      default permissions

    constraints are handled by specific hooks
    """
    entity = None  # make pylint happy

    def precommit_event(self):
        cnx = self.cnx
        entity = self.entity
        # update the in-memory schema first
        rdefdef, rdef = self.init_rdef(composite=entity.composite)
        # then make necessary changes to the system source database
        schema = cnx.vreg.schema
        rtype = rdefdef.name
        rschema = schema.rschema(rtype)
        # this have to be done before permissions setting
        if rschema.inlined:
            # need to add a column if the relation is inlined and if this is the
            # first occurence of "Subject relation Something" whatever Something
            if len(rschema.objects(rdefdef.subject)) == 1:
                add_inline_relation_column(cnx, rdefdef.subject, rtype)
            eschema = schema[rdefdef.subject]
            insert_rdef_on_subclasses(cnx, eschema, rschema, rdefdef,
                                      {'composite': entity.composite})
        else:
            if rschema.symmetric:
                # for symmetric relations, rdefs will store relation definitions
                # in both ways (i.e. (subj -> obj) and (obj -> subj))
                relation_already_defined = len(rschema.rdefs) > 2
            else:
                relation_already_defined = len(rschema.rdefs) > 1
            # need to create the relation if no relation definition in the
            # schema and if it has not been added during other event of the same
            # transaction
            if not (relation_already_defined or
                    rtype in cnx.transaction_data.get('createdtables', ())):
                rschema = schema.rschema(rtype)
                # create the necessary table
                for sql in y2sql.rschema2sql(rschema):
                    cnx.system_sql(sql)
                cnx.transaction_data.setdefault('createdtables', []).append(
                    rtype)

    # XXX revertprecommit_event


class RDefDelOp(MemSchemaOperation):
    """an actual relation has been removed"""
    rdef = None  # make pylint happy

    def precommit_event(self):
        cnx = self.cnx
        rdef = self.rdef
        rschema = rdef.rtype
        # make necessary changes to the system source database first
        rdeftype = rschema.final and 'CWAttribute' or 'CWRelation'
        execute = cnx.execute
        rset = execute('Any COUNT(X) WHERE X is %s, X relation_type R,'
                       'R eid %%(x)s' % rdeftype, {'x': rschema.eid})
        lastrel = rset[0][0] == 0
        # we have to update physical schema systematically for final and inlined
        # relations, but only if it's the last instance for this relation type
        # for other relations
        if (rschema.final or rschema.inlined):
            if not cnx.deleted_in_transaction(rdef.subject.eid):
                rset = execute('Any COUNT(X) WHERE X is %s, X relation_type R, '
                               'R eid %%(r)s, X from_entity E, E eid %%(e)s'
                               % rdeftype,
                               {'r': rschema.eid, 'e': rdef.subject.eid})
                if rset[0][0] == 0:
                    ptypes = cnx.transaction_data.setdefault('pendingrtypes', set())
                    ptypes.add(rschema.type)
                    DropColumn.get_instance(cnx).add_data((str(rdef.subject), str(rschema)))
                elif rschema.inlined:
                    cnx.system_sql('UPDATE %s%s SET %s%s=NULL WHERE '
                                   'EXISTS(SELECT 1 FROM entities '
                                   '       WHERE eid=%s%s AND type=%%(to_etype)s)'
                                   % (SQL_PREFIX, rdef.subject, SQL_PREFIX, rdef.rtype,
                                      SQL_PREFIX, rdef.rtype),
                                   {'to_etype': rdef.object.type})
        elif lastrel:
            DropRelationTable(cnx, str(rschema))
        else:
            cnx.system_sql('DELETE FROM %s_relation WHERE '
                           'EXISTS(SELECT 1 FROM entities '
                           '       WHERE eid=eid_from AND type=%%(from_etype)s)'
                           ' AND EXISTS(SELECT 1 FROM entities '
                           '       WHERE eid=eid_to AND type=%%(to_etype)s)'
                           % rschema,
                           {'from_etype': rdef.subject.type, 'to_etype': rdef.object.type})
        # then update the in-memory schema
        if rdef.subject not in ETYPE_NAME_MAP and rdef.object not in ETYPE_NAME_MAP:
            rschema.del_relation_def(rdef.subject, rdef.object)
        # if this is the last relation definition of this type, drop associated
        # relation type
        if lastrel and not cnx.deleted_in_transaction(rschema.eid):
            execute('DELETE CWRType X WHERE X eid %(x)s', {'x': rschema.eid})

    def revertprecommit_event(self):
        # revert changes on in memory schema
        #
        # Note: add_relation_def takes a RelationDefinition, not a
        # RelationDefinitionSchema, needs to fake it
        rdef = self.rdef
        rdef.name = str(rdef.rtype)
        if rdef.subject not in ETYPE_NAME_MAP and rdef.object not in ETYPE_NAME_MAP:
            self.cnx.vreg.schema.add_relation_def(rdef)


class RDefUpdateOp(MemSchemaOperation):
    """actually update some properties of a relation definition"""
    rschema = rdefkey = values = None  # make pylint happy
    rdef = oldvalues = None
    indexed_changed = null_allowed_changed = False

    def precommit_event(self):
        cnx = self.cnx
        rdef = self.rdef = self.rschema.rdefs[self.rdefkey]
        # update the in-memory schema first
        self.oldvalues = dict((attr, getattr(rdef, attr)) for attr in self.values)
        rdef.update(self.values)
        # then make necessary changes to the system source database
        syssource = cnx.repo.system_source
        if 'indexed' in self.values:
            syssource.update_rdef_indexed(cnx, rdef)
            self.indexed_changed = True
        if ('cardinality' in self.values and rdef.rtype.final
                and self.values['cardinality'][0] != self.oldvalues['cardinality'][0]):
            syssource.update_rdef_null_allowed(self.cnx, rdef)
            self.null_allowed_changed = True
        if 'fulltextindexed' in self.values:
            UpdateFTIndexOp.get_instance(cnx).add_data(rdef.subject)
        if 'formula' in self.values:
            RecomputeAttributeOperation.get_instance(cnx).add_data(rdef)

    def revertprecommit_event(self):
        if self.rdef is None:
            return
        # revert changes on in memory schema
        self.rdef.update(self.oldvalues)
        # revert changes on database
        syssource = self.cnx.repo.system_source
        if self.indexed_changed:
            syssource.update_rdef_indexed(self.cnx, self.rdef)
        if self.null_allowed_changed:
            syssource.update_rdef_null_allowed(self.cnx, self.rdef)


def _set_modifiable_constraints(rdef):
    # for proper in-place modification of in-memory schema: if rdef.constraints
    # is already a list, reuse it (we're updating multiple constraints of the
    # same rdef in the same transaction)
    if not isinstance(rdef.constraints, list):
        rdef.constraints = list(rdef.constraints)


class CWConstraintDelOp(MemSchemaOperation):
    """actually remove a constraint of a relation definition"""
    rdef = oldcstr = newcstr = None  # make pylint happy
    size_cstr_changed = unique_changed = False

    def precommit_event(self):
        cnx = self.cnx
        rdef = self.rdef
        # in-place modification of in-memory schema first
        _set_modifiable_constraints(rdef)
        if self.oldcstr in rdef.constraints:
            rdef.constraints.remove(self.oldcstr)
        else:
            self.critical('constraint %s for rdef %s was missing or already removed',
                          self.oldcstr, rdef)
        if cnx.deleted_in_transaction(rdef.eid):
            # don't try to alter a table that's going away (or is already gone)
            return
        # then update database: alter the physical schema on size/unique
        # constraint changes
        syssource = cnx.repo.system_source
        cstrtype = self.oldcstr.type()
        if cstrtype == 'SizeConstraint':
            # if the size constraint is being replaced with a new max size, we'll
            # call update_rdef_column in CWConstraintAddOp, skip it here
            for cstr in cnx.transaction_data.get('newsizecstr', ()):
                rdefentity = cstr.reverse_constrained_by[0]
                cstrrdef = cnx.vreg.schema.schema_by_eid(rdefentity.eid)
                if cstrrdef == rdef:
                    return

            # we found that the size constraint for this rdef is really gone,
            # not just replaced by another
            syssource.update_rdef_column(cnx, rdef)
            self.size_cstr_changed = True
        elif cstrtype == 'UniqueConstraint':
            syssource.update_rdef_unique(cnx, rdef)
            self.unique_changed = True
        elif cstrtype in ('BoundaryConstraint',
                          'IntervalBoundConstraint',
                          'StaticVocabularyConstraint'):
            cnx.system_sql('ALTER TABLE %s%s DROP CONSTRAINT %s'
                           % (SQL_PREFIX, rdef.subject, self.oldcstr.name_for(rdef)))

    def revertprecommit_event(self):
        # revert changes on in memory schema
        if self.newcstr is not None:
            self.rdef.constraints.remove(self.newcstr)
        if self.oldcstr is not None:
            self.rdef.constraints.append(self.oldcstr)
        # revert changes on database
        syssource = self.cnx.repo.system_source
        if self.size_cstr_changed:
            syssource.update_rdef_column(self.cnx, self.rdef)
        if self.unique_changed:
            syssource.update_rdef_unique(self.cnx, self.rdef)


class CWConstraintAddOp(hook.LateOperation, CWConstraintDelOp):
    """actually update constraint of a relation definition"""
    entity = None  # make pylint happy

    def precommit_event(self):
        cnx = self.cnx
        rdefentity = self.entity.reverse_constrained_by[0]
        rdef = self.rdef = cnx.vreg.schema.schema_by_eid(rdefentity.eid)
        cstrtype = self.entity.type
        if cstrtype in UNIQUE_CONSTRAINTS:
            oldcstr = self.oldcstr = rdef.constraint_by_type(cstrtype)
        else:
            oldcstr = None
        newcstr = self.newcstr = CONSTRAINTS[cstrtype].deserialize(self.entity.value)
        # in-place modification of in-memory schema first
        _set_modifiable_constraints(rdef)
        newcstr.eid = self.entity.eid
        if oldcstr is not None:
            rdef.constraints.remove(oldcstr)
        rdef.constraints.append(newcstr)
        # then update database: alter the physical schema on size/unique
        # constraint changes
        syssource = cnx.repo.system_source
        if cstrtype == 'SizeConstraint' and (oldcstr is None or
                                             oldcstr.max != newcstr.max):
            syssource.update_rdef_column(cnx, rdef)
            self.size_cstr_changed = True
        elif cstrtype == 'UniqueConstraint' and oldcstr is None:
            syssource.update_rdef_unique(cnx, rdef)
            self.unique_changed = True
        if cstrtype in ('BoundaryConstraint',
                        'IntervalBoundConstraint',
                        'StaticVocabularyConstraint'):
            cstrname, check = y2sql.check_constraint(rdef, newcstr, syssource.dbhelper,
                                                     prefix=SQL_PREFIX)
            # oldcstr is the new constraint when the attribute is being added in the same
            # transaction or when constraint value is updated. So we've to take care...
            if oldcstr is not None:
                oldcstrname = self.oldcstr.name_for(rdef)
                if oldcstrname != cstrname:
                    cnx.system_sql('ALTER TABLE %s%s DROP CONSTRAINT %s'
                                   % (SQL_PREFIX, rdef.subject, oldcstrname))
            cnx.system_sql('ALTER TABLE %s%s ADD CONSTRAINT %s CHECK(%s)' %
                           (SQL_PREFIX, rdef.subject, cstrname, check))


class CWUniqueTogetherConstraintAddOp(MemSchemaOperation):
    entity = None  # make pylint happy

    def precommit_event(self):
        cnx = self.cnx
        prefix = SQL_PREFIX
        entity = self.entity
        table = '%s%s' % (prefix, entity.constraint_of[0].name)
        cols = ['%s%s' % (prefix, r.name) for r in entity.relations]
        dbhelper = cnx.repo.system_source.dbhelper
        sqls = dbhelper.sqls_create_multicol_unique_index(table, cols, entity.name)
        for sql in sqls:
            cnx.system_sql(sql)

    def postcommit_event(self):
        entity = self.entity
        eschema = self.cnx.vreg.schema.schema_by_eid(entity.constraint_of[0].eid)
        attrs = [r.name for r in entity.relations]
        eschema._unique_together.append(attrs)


class CWUniqueTogetherConstraintDelOp(MemSchemaOperation):
    entity = cstrname = None  # make pylint happy
    cols = ()  # make pylint happy

    def insert_index(self):
        # We need to run before CWConstraintDelOp: if a size constraint is
        # removed and the column is part of a unique_together constraint, we
        # remove the unique_together index before changing the column's type.
        # SQL Server does not support unique indices on unlimited text columns.
        return 0

    def precommit_event(self):
        cnx = self.cnx
        prefix = SQL_PREFIX
        table = '%s%s' % (prefix, self.entity.type)
        dbhelper = cnx.repo.system_source.dbhelper
        cols = ['%s%s' % (prefix, c) for c in self.cols]
        sqls = dbhelper.sqls_drop_multicol_unique_index(table, cols, self.cstrname)
        for sql in sqls:
            cnx.system_sql(sql)

    def postcommit_event(self):
        eschema = self.cnx.vreg.schema.schema_by_eid(self.entity.eid)
        cols = set(self.cols)
        unique_together = [ut for ut in eschema._unique_together
                           if set(ut) != cols]
        eschema._unique_together = unique_together


# operations for in-memory schema synchronization  #############################

class MemSchemaCWETypeDel(MemSchemaOperation):
    """actually remove the entity type from the instance's schema"""
    etype = None  # make pylint happy

    def postcommit_event(self):
        # del_entity_type also removes entity's relations
        self.cnx.vreg.schema.del_entity_type(self.etype)


class MemSchemaCWRTypeAdd(MemSchemaOperation):
    """Revert addition of the relation type from the instance's schema if something goes wrong.
    """
    rtypedef = None  # make pylint happy

    def revertprecommit_event(self):
        self.cnx.vreg.schema.del_relation_type(self.rtypedef.name)


class MemSchemaCWRTypeDel(MemSchemaOperation):
    """actually remove the relation type from the instance's schema"""
    rtype = None  # make pylint happy

    def postcommit_event(self):
        try:
            self.cnx.vreg.schema.del_relation_type(self.rtype)
        except KeyError:
            # s/o entity type have already been deleted
            pass


class MemSchemaPermissionAdd(MemSchemaOperation):
    """synchronize schema when a *_permission relation has been added on a group
    """
    eid = action = group_eid = expr = None  # make pylint happy

    def precommit_event(self):
        """the observed connections.cnxset has been commited"""
        try:
            erschema = self.cnx.vreg.schema.schema_by_eid(self.eid)
        except KeyError:
            # duh, schema not found, log error and skip operation
            self.warning('no schema for %s', self.eid)
            return
        perms = list(erschema.action_permissions(self.action))
        if self.group_eid is not None:
            perm = self.cnx.entity_from_eid(self.group_eid).name
        else:
            perm = erschema.rql_expression(self.expr)
        try:
            perms.index(perm)
            self.warning('%s already in permissions for %s on %s',
                         perm, self.action, erschema)
        except ValueError:
            perms.append(perm)
            erschema.set_action_permissions(self.action, perms)

    # XXX revertprecommit_event


class MemSchemaPermissionDel(MemSchemaPermissionAdd):
    """synchronize schema when a *_permission relation has been deleted from a
    group
    """

    def precommit_event(self):
        """the observed connections set has been commited"""
        try:
            erschema = self.cnx.vreg.schema.schema_by_eid(self.eid)
        except KeyError:
            # duh, schema not found, log error and skip operation
            self.warning('no schema for %s', self.eid)
            return
        perms = list(erschema.action_permissions(self.action))
        if self.group_eid is not None:
            perm = self.cnx.entity_from_eid(self.group_eid).name
        else:
            perm = erschema.rql_expression(self.expr)
        try:
            perms.remove(perm)
            erschema.set_action_permissions(self.action, perms)
        except ValueError:
            self.error('can\'t remove permission %s for %s on %s',
                       perm, self.action, erschema)

    # XXX revertprecommit_event


class MemSchemaSpecializesAdd(MemSchemaOperation):
    etypeeid = parentetypeeid = None  # make pylint happy

    def precommit_event(self):
        eschema = self.cnx.vreg.schema.schema_by_eid(self.etypeeid)
        parenteschema = self.cnx.vreg.schema.schema_by_eid(self.parentetypeeid)
        eschema._specialized_type = parenteschema.type
        parenteschema._specialized_by.append(eschema.type)

    # XXX revertprecommit_event


class MemSchemaSpecializesDel(MemSchemaOperation):
    etypeeid = parentetypeeid = None  # make pylint happy

    def precommit_event(self):
        try:
            eschema = self.cnx.vreg.schema.schema_by_eid(self.etypeeid)
            parenteschema = self.cnx.vreg.schema.schema_by_eid(self.parentetypeeid)
        except KeyError:
            # etype removed, nothing to do
            return
        eschema._specialized_type = None
        parenteschema._specialized_by.remove(eschema.type)

    # XXX revertprecommit_event


# CWEType hooks ################################################################

class DelCWETypeHook(SyncSchemaHook):
    """before deleting a CWEType entity:
    * check that we don't remove a core entity type
    * cascade to delete related CWAttribute and CWRelation entities
    * instantiate an operation to delete the entity type on commit
    """
    __regid__ = 'syncdelcwetype'
    __select__ = SyncSchemaHook.__select__ & is_instance('CWEType')
    events = ('before_delete_entity',)

    def __call__(self):
        # final entities can't be deleted, don't care about that
        name = self.entity.name
        if name in CORE_TYPES:
            raise validation_error(self.entity, {None: _("can't be deleted")})
        # delete every entities of this type
        if name not in ETYPE_NAME_MAP:
            MemSchemaCWETypeDel(self._cw, etype=name)
        if not self.entity.final:
            DropTable(self._cw, table=SQL_PREFIX + name)


class AfterDelCWETypeHook(DelCWETypeHook):
    __regid__ = 'wfcleanup'
    events = ('after_delete_entity',)

    def __call__(self):
        # workflow cleanup
        self._cw.execute('DELETE Workflow X WHERE NOT X workflow_of Y')


class AfterAddCWETypeHook(DelCWETypeHook):
    """after adding a CWEType entity:
    * create the necessary table
    * set creation_date and modification_date by creating the necessary
      CWAttribute entities
    * add owned_by relation by creating the necessary CWRelation entity
    * register an operation to add the entity type to the instance's
      schema on commit
    """
    __regid__ = 'syncaddcwetype'
    events = ('after_add_entity',)

    def __call__(self):
        entity = self.entity
        if entity.cw_edited.get('final'):
            # final entity types don't need a table in the database and are
            # systematically added by yams at schema initialization time so
            # there is no need to do further processing. Simply assign its eid.
            self._cw.vreg.schema[entity.name].eid = entity.eid
            return
        CWETypeAddOp(self._cw, entity=entity)


class BeforeUpdateCWETypeHook(DelCWETypeHook):
    """check name change, handle final"""
    __regid__ = 'syncupdatecwetype'
    events = ('before_update_entity',)

    def __call__(self):
        entity = self.entity
        check_valid_changes(self._cw, entity, ro_attrs=('final',))
        # don't use getattr(entity, attr), we would get the modified value if any
        if 'name' in entity.cw_edited:
            oldname, newname = entity.cw_edited.oldnewvalue('name')
            if newname.lower() != oldname.lower():
                CWETypeRenameOp(self._cw, oldname=oldname, newname=newname)


# CWRType hooks ################################################################

class DelCWRTypeHook(SyncSchemaHook):
    """before deleting a CWRType entity:
    * check that we don't remove a core relation type
    * cascade to delete related CWAttribute and CWRelation entities
    * instantiate an operation to delete the relation type on commit
    """
    __regid__ = 'syncdelcwrtype'
    __select__ = SyncSchemaHook.__select__ & is_instance('CWRType')
    events = ('before_delete_entity',)

    def __call__(self):
        name = self.entity.name
        if name in CORE_TYPES:
            raise validation_error(self.entity, {None: _("can't be deleted")})
        # delete relation definitions using this relation type
        self._cw.execute('DELETE CWAttribute X WHERE X relation_type Y, Y eid %(x)s',
                         {'x': self.entity.eid})
        self._cw.execute('DELETE CWRelation X WHERE X relation_type Y, Y eid %(x)s',
                         {'x': self.entity.eid})
        MemSchemaCWRTypeDel(self._cw, rtype=name)


class AfterAddCWRTypeHook(SyncSchemaHook):
    """After a CWRType entity has been added, register an operation to add the
    relation type to the instance's schema on commit.

    We don't know yet at this point if a table is necessary, it will depend on
    further addition of relation definitions.
    """
    __regid__ = 'syncaddcwrtype'
    __select__ = SyncSchemaHook.__select__ & is_instance('CWRType')
    events = ('after_add_entity',)

    def __call__(self):
        rtypedef = self.rtype_def()
        # modify the instance's schema now since we'll usually need the type definition to do
        # further thing (e.g. add relation def of this type) but register and operation to revert
        # this if necessary
        self._cw.vreg.schema.add_relation_type(rtypedef)
        MemSchemaCWRTypeAdd(self._cw, rtypedef=rtypedef)

    def rtype_def(self):
        entity = self.entity
        return ybo.RelationType(name=entity.name,
                                description=entity.description,
                                inlined=entity.cw_edited.get('inlined', False),
                                symmetric=entity.cw_edited.get('symmetric', False),
                                eid=entity.eid)


class AfterAddCWComputedRTypeHook(AfterAddCWRTypeHook):
    """After a CWComputedRType entity has been added, register an operation to
    add the relation type to the instance's schema on commit.
    """
    __select__ = SyncSchemaHook.__select__ & is_instance('CWComputedRType')

    def rtype_def(self):
        entity = self.entity
        return ybo.ComputedRelation(name=entity.name,
                                    eid=entity.eid,
                                    rule=entity.rule)


class BeforeUpdateCWRTypeHook(SyncSchemaHook):
    """check name change, handle final"""
    __regid__ = 'syncupdatecwrtype'
    __select__ = SyncSchemaHook.__select__ & is_instance('CWRType')
    events = ('before_update_entity',)

    def __call__(self):
        entity = self.entity
        check_valid_changes(self._cw, entity)
        newvalues = {}
        for prop in ('symmetric', 'inlined', 'fulltext_container'):
            if prop in entity.cw_edited:
                old, new = entity.cw_edited.oldnewvalue(prop)
                if old != new:
                    newvalues[prop] = new
        if newvalues:
            rschema = self._cw.vreg.schema.rschema(entity.name)
            CWRTypeUpdateOp(self._cw, rschema=rschema, entity=entity,
                            values=newvalues)


class BeforeUpdateCWComputedRTypeHook(SyncSchemaHook):
    """check name change, handle final"""
    __regid__ = 'syncupdatecwcomputedrtype'
    __select__ = SyncSchemaHook.__select__ & is_instance('CWComputedRType')
    events = ('before_update_entity',)

    def __call__(self):
        entity = self.entity
        check_valid_changes(self._cw, entity)
        if 'rule' in entity.cw_edited:
            old, new = entity.cw_edited.oldnewvalue('rule')
            if old != new:
                rschema = self._cw.vreg.schema.rschema(entity.name)
                CWComputedRTypeUpdateOp(self._cw, rschema=rschema,
                                        entity=entity, rule=new)


class AfterDelRelationTypeHook(SyncSchemaHook):
    """before deleting a CWAttribute or CWRelation entity:
    * if this is a final or inlined relation definition, instantiate an
      operation to drop necessary column, else if this is the last instance
      of a non final relation, instantiate an operation to drop necessary
      table
    * instantiate an operation to delete the relation definition on commit
    * delete the associated relation type when necessary
    """
    __regid__ = 'syncdelrelationtype'
    __select__ = SyncSchemaHook.__select__ & hook.match_rtype('relation_type')
    events = ('after_delete_relation',)

    def __call__(self):
        cnx = self._cw
        try:
            rdef = cnx.vreg.schema.schema_by_eid(self.eidfrom)
        except KeyError:
            self.critical('cant get schema rdef associated to %s', self.eidfrom)
            return
        subjschema, rschema, objschema = rdef.as_triple()
        pendingrdefs = cnx.transaction_data.setdefault('pendingrdefs', set())
        # first delete existing relation if necessary
        if rschema.final:
            pendingrdefs.add((subjschema, rschema))
        else:
            pendingrdefs.add((subjschema, rschema, objschema))
        RDefDelOp(cnx, rdef=rdef)


# CWComputedRType hooks #######################################################

class DelCWComputedRTypeHook(SyncSchemaHook):
    """before deleting a CWComputedRType entity:
    * check that we don't remove a core relation type
    * instantiate an operation to delete the relation type on commit
    """
    __regid__ = 'syncdelcwcomputedrtype'
    __select__ = SyncSchemaHook.__select__ & is_instance('CWComputedRType')
    events = ('before_delete_entity',)

    def __call__(self):
        name = self.entity.name
        if name in CORE_TYPES:
            raise validation_error(self.entity, {None: _("can't be deleted")})
        MemSchemaCWRTypeDel(self._cw, rtype=name)


# CWAttribute / CWRelation hooks ###############################################

class AfterAddCWAttributeHook(SyncSchemaHook):
    __regid__ = 'syncaddcwattribute'
    __select__ = SyncSchemaHook.__select__ & is_instance('CWAttribute')
    events = ('after_add_entity',)

    def __call__(self):
        CWAttributeAddOp(self._cw, entity=self.entity)


class AfterAddCWRelationHook(AfterAddCWAttributeHook):
    __regid__ = 'syncaddcwrelation'
    __select__ = SyncSchemaHook.__select__ & is_instance('CWRelation')

    def __call__(self):
        CWRelationAddOp(self._cw, entity=self.entity)


class AfterUpdateCWRDefHook(SyncSchemaHook):
    __regid__ = 'syncaddcwattribute'
    __select__ = SyncSchemaHook.__select__ & is_instance('CWAttribute',
                                                         'CWRelation')
    events = ('before_update_entity',)

    def __call__(self):
        entity = self.entity
        if self._cw.deleted_in_transaction(entity.eid):
            return
        subjtype = entity.stype.name
        objtype = entity.otype.name
        if subjtype in ETYPE_NAME_MAP or objtype in ETYPE_NAME_MAP:
            return
        rschema = self._cw.vreg.schema[entity.rtype.name]
        # note: do not access schema rdef here, it may be added later by an
        # operation
        newvalues = {}
        for prop in RelationDefinitionSchema.rproperty_defs(objtype):
            if prop == 'constraints':
                continue
            if prop == 'order':
                attr = 'ordernum'
            else:
                attr = prop
            if attr in entity.cw_edited:
                old, new = entity.cw_edited.oldnewvalue(attr)
                if old != new:
                    newvalues[prop] = new
        if newvalues:
            RDefUpdateOp(self._cw, rschema=rschema, rdefkey=(subjtype, objtype),
                         values=newvalues)


# constraints synchronization hooks ############################################

class AfterAddCWConstraintHook(SyncSchemaHook):
    __regid__ = 'syncaddcwconstraint'
    __select__ = SyncSchemaHook.__select__ & is_instance('CWConstraint')
    events = ('after_add_entity', 'after_update_entity')

    def __call__(self):
        if self.entity.cstrtype[0].name == 'SizeConstraint':
            txdata = self._cw.transaction_data
            if 'newsizecstr' not in txdata:
                txdata['newsizecstr'] = set()
            txdata['newsizecstr'].add(self.entity)
        CWConstraintAddOp(self._cw, entity=self.entity)


class AfterAddConstrainedByHook(SyncSchemaHook):
    __regid__ = 'syncaddconstrainedby'
    __select__ = SyncSchemaHook.__select__ & hook.match_rtype('constrained_by')
    events = ('after_add_relation',)

    def __call__(self):
        if self._cw.added_in_transaction(self.eidfrom):
            # used by get_constraints() which is called in CWAttributeAddOp
            self._cw.transaction_data.setdefault(self.eidfrom, []).append(self.eidto)


class BeforeDeleteCWConstraintHook(SyncSchemaHook):
    __regid__ = 'syncdelcwconstraint'
    __select__ = SyncSchemaHook.__select__ & is_instance('CWConstraint')
    events = ('before_delete_entity',)

    def __call__(self):
        entity = self.entity
        schema = self._cw.vreg.schema
        try:
            # KeyError, e.g. composite chain deletion
            rdef = schema.schema_by_eid(entity.reverse_constrained_by[0].eid)
            # IndexError
            cstr = rdef.constraint_by_eid(entity.eid)
        except (KeyError, IndexError):
            self._cw.critical('constraint type no more accessible')
        else:
            CWConstraintDelOp(self._cw, rdef=rdef, oldcstr=cstr)


# unique_together constraints
# XXX: use setoperations and before_add_relation here (on constraint_of and relations)
class AfterAddCWUniqueTogetherConstraintHook(SyncSchemaHook):
    __regid__ = 'syncadd_cwuniquetogether_constraint'
    __select__ = SyncSchemaHook.__select__ & is_instance('CWUniqueTogetherConstraint')
    events = ('after_add_entity',)

    def __call__(self):
        CWUniqueTogetherConstraintAddOp(self._cw, entity=self.entity)


class BeforeDeleteConstraintOfHook(SyncSchemaHook):
    __regid__ = 'syncdelconstraintof'
    __select__ = SyncSchemaHook.__select__ & hook.match_rtype('constraint_of')
    events = ('before_delete_relation',)

    def __call__(self):
        if self._cw.deleted_in_transaction(self.eidto):
            return
        schema = self._cw.vreg.schema
        cstr = self._cw.entity_from_eid(self.eidfrom)
        entity = schema.schema_by_eid(self.eidto)
        cols = tuple(r.name for r in cstr.relations)
        CWUniqueTogetherConstraintDelOp(self._cw, entity=entity,
                                        cstrname=cstr.name, cols=cols)


# permissions synchronization hooks ############################################

class AfterAddPermissionHook(SyncSchemaHook):
    """added entity/relation *_permission, need to update schema"""
    __regid__ = 'syncaddperm'
    __select__ = SyncSchemaHook.__select__ & hook.match_rtype(
        'read_permission', 'add_permission', 'delete_permission',
        'update_permission')
    events = ('after_add_relation',)

    def __call__(self):
        action = self.rtype.split('_', 1)[0]
        if self._cw.entity_type(self.eidto) == 'CWGroup':
            MemSchemaPermissionAdd(self._cw, action=action, eid=self.eidfrom,
                                   group_eid=self.eidto)
        else:  # RQLExpression
            expr = self._cw.entity_from_eid(self.eidto).expression
            MemSchemaPermissionAdd(self._cw, action=action, eid=self.eidfrom,
                                   expr=expr)


class BeforeDelPermissionHook(AfterAddPermissionHook):
    """delete entity/relation *_permission, need to update schema

    skip the operation if the related type is being deleted
    """
    __regid__ = 'syncdelperm'
    events = ('before_delete_relation',)

    def __call__(self):
        if self._cw.deleted_in_transaction(self.eidfrom):
            return
        action = self.rtype.split('_', 1)[0]
        if self._cw.entity_type(self.eidto) == 'CWGroup':
            MemSchemaPermissionDel(self._cw, action=action, eid=self.eidfrom,
                                   group_eid=self.eidto)
        else:  # RQLExpression
            expr = self._cw.entity_from_eid(self.eidto).expression
            MemSchemaPermissionDel(self._cw, action=action, eid=self.eidfrom,
                                   expr=expr)


class UpdateFTIndexOp(hook.DataOperationMixIn, hook.SingleLastOperation):
    """operation to update full text indexation of entity whose schema change

    We wait after the commit to as the schema in memory is only updated after
    the commit.
    """
    containercls = list

    def postcommit_event(self):
        cnx = self.cnx
        source = cnx.repo.system_source
        schema = cnx.repo.vreg.schema
        to_reindex = self.get_data()
        self.info('%i etypes need full text indexed reindexation',
                  len(to_reindex))
        for etype in to_reindex:
            rset = cnx.execute('Any X WHERE X is %s' % etype)
            self.info('Reindexing full text index for %i entity of type %s',
                      len(rset), etype)
            still_fti = list(schema[etype].indexable_attributes())
            for entity in rset.entities():
                source.fti_unindex_entities(cnx, [entity])
                for container in entity.cw_adapt_to('IFTIndexable').fti_containers():
                    if still_fti or container is not entity:
                        source.fti_unindex_entities(cnx, [container])
                        source.fti_index_entities(cnx, [container])
        if to_reindex:
            # Transaction has already been committed
            cnx.cnxset.commit()


# specializes synchronization hooks ############################################

class AfterAddSpecializesHook(SyncSchemaHook):
    __regid__ = 'syncaddspecializes'
    __select__ = SyncSchemaHook.__select__ & hook.match_rtype('specializes')
    events = ('after_add_relation',)

    def __call__(self):
        MemSchemaSpecializesAdd(self._cw, etypeeid=self.eidfrom,
                                parentetypeeid=self.eidto)


class AfterDelSpecializesHook(SyncSchemaHook):
    __regid__ = 'syncdelspecializes'
    __select__ = SyncSchemaHook.__select__ & hook.match_rtype('specializes')
    events = ('after_delete_relation',)

    def __call__(self):
        MemSchemaSpecializesDel(self._cw, etypeeid=self.eidfrom,
                                parentetypeeid=self.eidto)



from six.moves import http_client

from logilab.common.testlib import Tags
from cubicweb.devtools.webtest import CubicWebTestTC


class CWTTC(CubicWebTestTC):
    def test_response(self):
        response = self.webapp.get('/')
        self.assertEqual(200, response.status_int)

    def test_base_url(self):
        if self.config['base-url'] not in self.webapp.get('/').text:
            self.fail('no mention of base url in retrieved page')


class CWTIdentTC(CubicWebTestTC):
    test_db_id = 'webtest-ident'
    anonymous_allowed = False
    tags = CubicWebTestTC.tags | Tags(('auth',))

    def test_reponse_denied(self):
        res = self.webapp.get('/', expect_errors=True)
        self.assertEqual(http_client.FORBIDDEN, res.status_int)

    def test_login(self):
        res = self.webapp.get('/', expect_errors=True)
        self.assertEqual(http_client.FORBIDDEN, res.status_int)

        self.login(self.admlogin, self.admpassword)
        res = self.webapp.get('/')
        self.assertEqual(http_client.OK, res.status_int)

        self.logout()
        res = self.webapp.get('/', expect_errors=True)
        self.assertEqual(http_client.FORBIDDEN, res.status_int)


if __name__ == '__main__':
    import unittest
    unittest.main()



# copyright 2003-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""html widgets

those are in cubicweb since we need to know available widgets at schema
serialization time
"""

import random
from math import floor

from six import add_metaclass
from six.moves import range

from logilab.mtconverter import xml_escape
from logilab.common.deprecation import class_deprecated

from cubicweb.utils import UStringIO
from cubicweb.uilib import toggle_action, htmlescape
from cubicweb.web import jsonize
from cubicweb.web.component import _bwcompatible_render_item

# XXX HTMLWidgets should have access to req (for datadir / static urls,
#     i18n strings, etc.)
class HTMLWidget(object):

    def _initialize_stream(self, w=None):
        if w:
            self.w = w
        else:
            self._stream = UStringIO()
            self.w = self._stream.write

    def _render(self):
        raise NotImplementedError

    def render(self, w=None):
        self._initialize_stream(w)
        self._render()
        if w is None:
            return self._stream.getvalue()

    def is_empty(self):
        return False


class BoxWidget(HTMLWidget): # XXX Deprecated

    def __init__(self, title, id, items=None, _class="boxFrame",
                 islist=True, shadow=True, escape=True):
        self.title = title
        self.id = id
        self.items = items or []
        self._class = _class
        self.islist = islist
        self.shadow = shadow
        self.escape = escape

    def __len__(self):
        return len(self.items)

    def is_empty(self):
        return len(self) == 0

    def append(self, item):
        self.items.append(item)

    def extend(self, items):
        self.items.extend(items)

    title_class = 'boxTitle'
    main_div_class = 'boxContent'
    listing_class = 'boxListing'

    def box_begin_content(self):
        self.w(u'<div class="%s">\n' % self.main_div_class)
        if self.islist:
            self.w(u'<ul class="%s">' % self.listing_class)

    def box_end_content(self):
        if self.islist:
            self.w(u'</ul>\n')
        self.w(u'</div>\n')
        if self.shadow:
            self.w(u'<div class="shadow">&#160;</div>')

    def _render(self):
        if self.id:
            self.w(u'<div class="%s" id="%s">' % (self._class, self.id))
        else:
            self.w(u'<div class="%s">' % self._class)
        if self.title:
            if self.escape:
                title = '<span>%s</span>' % xml_escape(self.title)
            else:
                title = '<span>%s</span>' % self.title
            self.w(u'<div class="%s">%s</div>' % (self.title_class, title))
        if self.items:
            self.box_begin_content()
            for item in self.items:
                _bwcompatible_render_item(self.w, item)
            self.box_end_content()
        self.w(u'</div>')


@add_metaclass(class_deprecated)
class SideBoxWidget(BoxWidget):
    """default CubicWeb's sidebox widget"""
    __deprecation_warning__ = '[3.10] class %(cls)s is deprecated'

    title_class = u'sideBoxTitle'
    main_div_class = u'sideBoxBody'
    listing_class = ''

    def __init__(self, title, id=None):
        super(SideBoxWidget, self).__init__(title, id=id, _class='sideBox',
                                            shadow=False)


class MenuWidget(BoxWidget):

    main_div_class = 'menuContent'
    listing_class = 'menuListing'

    def box_end_content(self):
        if self.islist:
            self.w(u'</ul>\n')
        self.w(u'</div>\n')


class RawBoxItem(HTMLWidget): # XXX deprecated
    """a simple box item displaying raw data"""

    def __init__(self, label, liclass=None):
        self.label = label
        self.liclass = liclass

    def _start_li(self):
        if self.liclass is None:
            return u'<li>'
        else:
            return u'<li class="%s">' % self.liclass

    def _render(self):
        self.w(u'%s%s</li>' % (self._start_li(), self.label))


class BoxMenu(RawBoxItem):
    """a menu in a box"""

    link_class = 'boxMenu'

    def __init__(self, label, items=None, isitem=True, liclass=None, ident=None,
                 link_class=None):
        super(BoxMenu, self).__init__(label, liclass)
        self.items = items or []
        self.isitem = isitem
        self.ident = ident or u'boxmenu_%s' % label.replace(' ', '_').replace("'", '')
        if link_class:
            self.link_class = link_class

    def append(self, item):
        self.items.append(item)

    def _begin_menu(self, ident):
        self.w(u'<ul id="%s" class="hidden">' % ident)

    def _end_menu(self):
        self.w(u'</ul>')

    def _render(self):
        if self.isitem:
            self.w(self._start_li())
        ident = self.ident
        self.w(u'<a href="%s" class="%s">%s</a>' % (
            toggle_action(ident), self.link_class, self.label))
        self._begin_menu(ident)
        for item in self.items:
            _bwcompatible_render_item(self.w, item)
        self._end_menu()
        if self.isitem:
            self.w(u'</li>')


class PopupBoxMenu(BoxMenu):
    """like BoxMenu but uses div and specific css class
    in order to behave like a popup menu
    """
    link_class = 'popupMenu'

    def _begin_menu(self, ident):
        self.w(u'<div class="popupWrapper"><div id="%s" class="hidden popup"><ul>' % ident)

    def _end_menu(self):
        self.w(u'</ul></div></div>')


@add_metaclass(class_deprecated)
class BoxField(HTMLWidget):
    """couples label / value meant to be displayed in a box"""
    __deprecation_warning__ = '[3.10] class %(cls)s is deprecated'
    def __init__(self, label, value):
        self.label = label
        self.value = value

    def _render(self):
        self.w(u'<li><div><span class="label">%s</span>&#160;'
               u'<span class="value">%s</span></div></li>'
               % (self.label, self.value))


@add_metaclass(class_deprecated)
class BoxSeparator(HTMLWidget):
    """a menu separator"""
    __deprecation_warning__ = '[3.10] class %(cls)s is deprecated'

    def _render(self):
        self.w(u'</ul><hr class="boxSeparator"/><ul>')


@add_metaclass(class_deprecated)
class BoxLink(HTMLWidget):
    """a link in a box"""
    __deprecation_warning__ = '[3.10] class %(cls)s is deprecated'
    def __init__(self, href, label, _class='', title='', ident='', escape=False):
        self.href = href
        if escape:
            self.label = xml_escape(label)
        else:
            self.label = label
        self._class = _class or ''
        self.title = title
        self.ident = ident

    def _render(self):
        link = u'<a href="%s" title="%s">%s</a>' % (
            xml_escape(self.href), xml_escape(self.title), self.label)
        if self.ident:
            self.w(u'<li id="%s" class="%s">%s</li>\n' % (self.ident, self._class, link))
        else:
            self.w(u'<li class="%s">%s</li>\n' % (self._class, link))


@add_metaclass(class_deprecated)
class BoxHtml(HTMLWidget):
    """a form in a box"""
    __deprecation_warning__ = '[3.10] class %(cls)s is deprecated'
    def __init__(self, rawhtml):
        self.rawhtml = rawhtml

    def _render(self):
        self.w(self.rawhtml)


class TableColumn(object):
    def __init__(self, name, rset_sortcol):
        """
        :param name: the column's name
        :param rset_sortcol: the model's column used to sort this column view
        """
        self.name = name
        self.cellrenderers = []
        self.rset_sortcol = rset_sortcol
        self.cell_attrs = {}

    def append_renderer(self, cellvid, colindex):
        # XXX (adim) : why do we need colindex here ?
        self.cellrenderers.append( (cellvid, colindex) )

    def add_attr(self, attr, value):
        self.cell_attrs[attr] = value


class SimpleTableModel(object):
    """
    uses a list of lists as a storage backend

    NB: the model expectes the cellvid passed to
    TableColumn.append_renderer to be a callable accepting a single
    argument and returning a unicode object
    """
    def __init__(self, rows):
        self._rows = rows

    def get_rows(self):
        return self._rows

    def render_cell(self, cellvid, rowindex, colindex, w):
        value = self._rows[rowindex][colindex]
        w(cellvid(value))

    @htmlescape
    @jsonize
    def sortvalue(self, rowindex, colindex):
        value =  self._rows[rowindex][colindex]
        if value is None:
            return u''
        elif isinstance(value, int):
            return u'%09d' % value
        else:
            return unicode(value)


class TableWidget(HTMLWidget):
    """
    Display data in a Table with sortable column.

    When using remember to include the required css and js with:

    self._cw.add_js('jquery.tablesorter.js')
    self._cw.add_css(('cubicweb.tablesorter.css', 'cubicweb.tableview.css'))
    """
    highlight = "onmouseover=\"$(this).addClass('highlighted');\" " \
                "onmouseout=\"$(this).removeClass('highlighted');\""

    def __init__(self, model):
        self.model = model
        self.columns = []

    def append_column(self, column):
        """
        :type column: TableColumn
        """
        self.columns.append(column)

    def _render(self):
        self.w(u'<table class="listing">')
        self.w(u'<thead>')
        self.w(u'<tr class="header">')
        for column in self.columns:
            attrs = ('%s="%s"' % (name, value) for name, value in column.cell_attrs.items())
            self.w(u'<th %s>%s</th>' % (' '.join(attrs), column.name or u''))
        self.w(u'</tr>')
        self.w(u'</thead><tbody>')
        for rowindex in range(len(self.model.get_rows())):
            klass = (rowindex%2==1) and 'odd' or 'even'
            self.w(u'<tr class="%s" %s>' % (klass, self.highlight))
            for column, sortvalue in self.itercols(rowindex):
                attrs = dict(column.cell_attrs)
                attrs["cubicweb:sortvalue"] = sortvalue
                attrs = ('%s="%s"' % (name, value) for name, value in attrs.items())
                self.w(u'<td %s>' % (' '.join(attrs)))
                for cellvid, colindex in column.cellrenderers:
                    self.model.render_cell(cellvid, rowindex, colindex, w=self.w)
                self.w(u'</td>')
            self.w(u'</tr>')
        self.w(u'</tbody>')
        self.w(u'</table>')

    def itercols(self, rowindex):
        for column in self.columns:
            yield column, self.model.sortvalue(rowindex, column.rset_sortcol)



# copyright 2003-2012 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.

from warnings import warn

from six import string_types

from logilab.common.deprecation import deprecated, class_renamed

from cubicweb.predicates import *


warn('[3.15] cubicweb.selectors renamed into cubicweb.predicates',
     DeprecationWarning, stacklevel=2)

# XXX pre 3.15 bw compat
from cubicweb.appobject import (objectify_selector, traced_selection,
                                lltrace, yes)

ExpectedValueSelector = class_renamed('ExpectedValueSelector',
                                      ExpectedValuePredicate)
EClassSelector = class_renamed('EClassSelector', EClassPredicate)
EntitySelector = class_renamed('EntitySelector', EntityPredicate)


class on_transition(is_in_state):
    """Return 1 if entity is in one of the transitions given as argument list

    Especially useful to match passed transition to enable notifications when
    your workflow allows several transition to the same states.

    Note that if workflow `change_state` adapter method is used, this predicate
    will not be triggered.

    You should use this instead of your own :class:`score_entity` predicate to
    avoid some gotchas:

    * possible views gives a fake entity with no state
    * you must use the latest tr info thru the workflow adapter for repository
      side checking of the current state

    In debug mode, this predicate can raise:
    :raises: :exc:`ValueError` for unknown transition names
        (etype workflow only not checked in custom workflow)

    :rtype: int
    """
    @deprecated('[3.12] on_transition is deprecated, you should rather use '
                'on_fire_transition(etype, trname)')
    def __init__(self, *expected):
        super(on_transition, self).__init__(*expected)

    def _score(self, adapted):
        trinfo = adapted.latest_trinfo()
        if trinfo and trinfo.by_transition:
            return trinfo.by_transition[0].name in self.expected

    def _validate(self, adapted):
        wf = adapted.current_workflow
        valid = [n.name for n in wf.reverse_transition_of]
        unknown = sorted(self.expected.difference(valid))
        if unknown:
            raise ValueError("%s: unknown transition(s): %s"
                             % (wf.name, ",".join(unknown)))


entity_implements = class_renamed('entity_implements', is_instance)

class _but_etype(EntityPredicate):
    """accept if the given entity types are not found in the result set.

    See `EntityPredicate` documentation for behaviour when row is not specified.

    :param *etypes: entity types (`string_types`) which should be refused
    """
    def __init__(self, *etypes):
        super(_but_etype, self).__init__()
        self.but_etypes = etypes

    def score(self, req, rset, row, col):
        if rset.description[row][col] in self.but_etypes:
            return 0
        return 1

but_etype = class_renamed('but_etype', _but_etype, 'use ~is_instance(*etypes) instead')

# XXX deprecated the one_* variants of predicates below w/ multi_xxx(nb=1)?
#     take care at the implementation though (looking for the 'row' argument's
#     value)
two_lines_rset = class_renamed('two_lines_rset', multi_lines_rset)
two_cols_rset = class_renamed('two_cols_rset', multi_columns_rset)
two_etypes_rset = class_renamed('two_etypes_rset', multi_etypes_rset)



from cubicweb.pyramid.test import PyramidCWTest

from cubicweb.view import View
from cubicweb.web import Redirect
from cubicweb import ValidationError


class Redirector(View):
    __regid__ = 'redirector'

    def call(self, rset=None):
        self._cw.set_header('Cache-Control', 'no-cache')
        raise Redirect('http://example.org')


def put_in_uncommitable_state(request):
    try:
        request.cw_cnx.execute('SET U login NULL WHERE U login "anon"')
    except ValidationError:
        pass
    request.response.body = b'OK'
    return request.response


class CoreTest(PyramidCWTest):
    anonymous_allowed = True

    def includeme(self, config):
        config.add_route('uncommitable', '/uncommitable')
        config.add_view(put_in_uncommitable_state, route_name='uncommitable')

    def test_cw_to_pyramid_copy_headers_on_redirect(self):
        self.vreg.register(Redirector)
        try:
            res = self.webapp.get('/?vid=redirector', expect_errors=True)
            self.assertEqual(res.status_int, 303)
            self.assertEqual(res.headers['Cache-Control'], 'no-cache')
        finally:
            self.vreg.unregister(Redirector)

    def test_uncommitable_cnx(self):
        res = self.webapp.get('/uncommitable')
        self.assertEqual(res.text, 'OK')
        self.assertEqual(res.status_int, 200)


if __name__ == '__main__':
    from unittest import main
    main()



# copyright 2003-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""Some utilities for CubicWeb server/clients."""

from __future__ import division

import base64
import decimal
import datetime
import random
import re
import json

from operator import itemgetter
from inspect import getargspec
from itertools import repeat
from uuid import uuid4
from warnings import warn
from threading import Lock
from logging import getLogger

from six import text_type

from logilab.mtconverter import xml_escape
from logilab.common.deprecation import deprecated
from logilab.common.date import ustrftime

from cubicweb import Binary


_MARKER = object()

# initialize random seed from current time
random.seed()


def admincnx(appid):
    from cubicweb import repoapi
    from cubicweb.cwconfig import CubicWebConfiguration
    from cubicweb.server.repository import Repository
    config = CubicWebConfiguration.config_for(appid)

    login = config.default_admin_config['login']
    password = config.default_admin_config['password']

    repo = Repository(config)
    repo.bootstrap()
    return repoapi.connect(repo, login, password=password)


def make_uid(key=None):
    """Return a unique identifier string.

    if specified, `key` is used to prefix the generated uid so it can be used
    for instance as a DOM id or as sql table name.

    See uuid.uuid4 documentation for the shape of the generated identifier, but
    this is basically a 32 bits hexadecimal string.
    """
    if key is None:
        return uuid4().hex
    return str(key) + uuid4().hex


def support_args(callable, *argnames):
    """return true if the callable support given argument names"""
    if isinstance(callable, type):
        callable = callable.__init__
    argspec = getargspec(callable)
    if argspec[2]:
        return True
    for argname in argnames:
        if argname not in argspec[0]:
            return False
    return True


class wrap_on_write(object):
    """ Sometimes it is convenient to NOT write some container element
    if it happens that there is nothing to be written within,
    but this cannot be known beforehand.
    Hence one can do this:

    .. sourcecode:: python

       with wrap_on_write(w, '<div class="foo">', '</div>') as wow:
           component.render_stuff(wow)
    """
    def __init__(self, w, tag, closetag=None):
        self.written = False
        self.tag = text_type(tag)
        self.closetag = closetag
        self.w = w

    def __enter__(self):
        return self

    def __call__(self, data):
        if self.written is False:
            self.w(self.tag)
            self.written = True
        self.w(data)

    def __exit__(self, exctype, value, traceback):
        if self.written is True:
            if self.closetag:
                self.w(text_type(self.closetag))
            else:
                self.w(self.tag.replace('<', '</', 1))


# use networkX instead ?
# http://networkx.lanl.gov/reference/algorithms.traversal.html#module-networkx.algorithms.traversal.astar
def transitive_closure_of(entity, rtype, _seen=None):
    """return transitive closure *for the subgraph starting from the given
    entity* (eg 'parent' entities are not included in the results)
    """
    if _seen is None:
        _seen = set()
    _seen.add(entity.eid)
    yield entity
    for child in getattr(entity, rtype):
        if child.eid in _seen:
            continue
        for subchild in transitive_closure_of(child, rtype, _seen):
            yield subchild


class RepeatList(object):
    """fake a list with the same element in each row"""
    __slots__ = ('_size', '_item')
    def __init__(self, size, item):
        self._size = size
        self._item = item
    def __repr__(self):
        return '<cubicweb.utils.RepeatList at %s item=%s size=%s>' % (
            id(self), self._item, self._size)
    def __len__(self):
        return self._size
    def __iter__(self):
        return repeat(self._item, self._size)
    def __getitem__(self, index):
        if isinstance(index, slice):
            # XXX could be more efficient, but do we bother?
            return ([self._item] * self._size)[index]
        return self._item
    def __delitem__(self, idc):
        assert self._size > 0
        self._size -= 1
    def __add__(self, other):
        if isinstance(other, RepeatList):
            if other._item == self._item:
                return RepeatList(self._size + other._size, self._item)
            return ([self._item] * self._size) + other[:]
        return ([self._item] * self._size) + other
    def __radd__(self, other):
        if isinstance(other, RepeatList):
            if other._item == self._item:
                return RepeatList(self._size + other._size, self._item)
            return other[:] + ([self._item] * self._size)
        return other[:] + ([self._item] * self._size)
    def __eq__(self, other):
        if isinstance(other, RepeatList):
            return other._size == self._size and other._item == self._item
        return self[:] == other
    def __ne__(self, other):
        return not (self == other)
    def __hash__(self):
        raise NotImplementedError
    def pop(self, i):
        self._size -= 1


class UStringIO(list):
    """a file wrapper which automatically encode unicode string to an encoding
    specifed in the constructor
    """

    def __init__(self, tracewrites=False, *args, **kwargs):
        self.tracewrites = tracewrites
        super(UStringIO, self).__init__(*args, **kwargs)

    def __bool__(self):
        return True

    __nonzero__ = __bool__

    def write(self, value):
        assert isinstance(value, text_type), u"unicode required not %s : %s"\
                                     % (type(value).__name__, repr(value))
        if self.tracewrites:
            from traceback import format_stack
            stack = format_stack(None)[:-1]
            escaped_stack = xml_escape(json_dumps(u'\n'.join(stack)))
            escaped_html = xml_escape(value).replace('\n', '<br/>\n')
            tpl = u'<span onclick="alert(%s)">%s</span>'
            value = tpl % (escaped_stack, escaped_html)
        self.append(value)

    def getvalue(self):
        return u''.join(self)

    def __repr__(self):
        return '<%s at %#x>' % (self.__class__.__name__, id(self))


class HTMLHead(UStringIO):
    """wraps HTML header's stream

    Request objects use a HTMLHead instance to ease adding of
    javascripts and stylesheets
    """
    js_unload_code = u'''if (typeof(pageDataUnloaded) == 'undefined') {
    jQuery(window).unload(unloadPageData);
    pageDataUnloaded = true;
}'''
    script_opening = u'<script type="text/javascript">\n'
    script_closing = u'\n</script>'

    def __init__(self, req, *args, **kwargs):
        super(HTMLHead, self).__init__(*args, **kwargs)
        self.jsvars = []
        self.jsfiles = []
        self.cssfiles = []
        self.ie_cssfiles = []
        self.post_inlined_scripts = []
        self.pagedata_unload = False
        self._cw = req
        self.datadir_url = req.datadir_url

    def add_raw(self, rawheader):
        self.write(rawheader)

    def define_var(self, var, value, override=True):
        """adds a javascript var declaration / assginment in the header

        :param var: the variable name
        :param value: the variable value (as a raw python value,
                      it will be jsonized later)
        :param override: if False, don't set the variable value if the variable
                         is already defined. Default is True.
        """
        self.jsvars.append( (var, value, override) )

    def add_post_inline_script(self, content):
        self.post_inlined_scripts.append(content)

    def add_onload(self, jscode):
        self.add_post_inline_script(u"""$(cw).one('server-response', function(event) {
%s});""" % jscode)


    def add_js(self, jsfile):
        """adds `jsfile` to the list of javascripts used in the webpage

        This function checks if the file has already been added
        :param jsfile: the script's URL
        """
        if jsfile not in self.jsfiles:
            self.jsfiles.append(jsfile)

    def add_css(self, cssfile, media='all'):
        """adds `cssfile` to the list of javascripts used in the webpage

        This function checks if the file has already been added
        :param cssfile: the stylesheet's URL
        """
        if (cssfile, media) not in self.cssfiles:
            self.cssfiles.append( (cssfile, media) )

    def add_ie_css(self, cssfile, media='all', iespec=u'[if lt IE 8]'):
        """registers some IE specific CSS"""
        if (cssfile, media, iespec) not in self.ie_cssfiles:
            self.ie_cssfiles.append( (cssfile, media, iespec) )

    def add_unload_pagedata(self):
        """registers onunload callback to clean page data on server"""
        if not self.pagedata_unload:
            self.post_inlined_scripts.append(self.js_unload_code)
            self.pagedata_unload = True

    def concat_urls(self, urls):
        """concatenates urls into one url usable by Apache mod_concat

        This method returns the url without modifying it if there is only
        one element in the list
        :param urls: list of local urls/filenames to concatenate
        """
        if len(urls) == 1:
            return urls[0]
        len_prefix = len(self.datadir_url)
        concated = u','.join(url[len_prefix:] for url in urls)
        return (u'%s??%s' % (self.datadir_url, concated))

    def group_urls(self, urls_spec):
        """parses urls_spec in order to generate concatenated urls
        for js and css includes

        This method checks if the file is local and if it shares options
        with direct neighbors
        :param urls_spec: entire list of urls/filenames to inspect
        """
        concatable = []
        prev_islocal = False
        prev_key = None
        for url, key in urls_spec:
            islocal = url.startswith(self.datadir_url)
            if concatable and (islocal != prev_islocal or key != prev_key):
                yield (self.concat_urls(concatable), prev_key)
                del concatable[:]
            if not islocal:
                yield (url, key)
            else:
                concatable.append(url)
            prev_islocal = islocal
            prev_key = key
        if concatable:
            yield (self.concat_urls(concatable), prev_key)


    def getvalue(self, skiphead=False):
        """reimplement getvalue to provide a consistent (and somewhat browser
        optimzed cf. http://stevesouders.com/cuzillion) order in external
        resources declaration
        """
        w = self.write
        # 1/ variable declaration if any
        if self.jsvars:
            if skiphead:
                w(u'<cubicweb:script>')
            else:
                w(self.script_opening)
            for var, value, override in self.jsvars:
                vardecl = u'%s = %s;' % (var, json.dumps(value))
                if not override:
                    vardecl = (u'if (typeof %s == "undefined") {%s}' %
                               (var, vardecl))
                w(vardecl + u'\n')
            if skiphead:
                w(u'</cubicweb:script>')
            else:
                w(self.script_closing)
        # 2/ css files
        ie_cssfiles = ((x, (y, z)) for x, y, z in self.ie_cssfiles)
        if self.datadir_url and self._cw.vreg.config['concat-resources']:
            cssfiles = self.group_urls(self.cssfiles)
            ie_cssfiles = self.group_urls(ie_cssfiles)
            jsfiles = (x for x, _ in self.group_urls((x, None) for x in self.jsfiles))
        else:
            cssfiles = self.cssfiles
            jsfiles = self.jsfiles
        for cssfile, media in cssfiles:
            w(u'<link rel="stylesheet" type="text/css" media="%s" href="%s"/>\n' %
              (media, xml_escape(cssfile)))
        # 3/ ie css if necessary
        if self.ie_cssfiles: # use self.ie_cssfiles because `ie_cssfiles` is a genexp
            for cssfile, (media, iespec) in ie_cssfiles:
                w(u'<!--%s>\n' % iespec)
                w(u'<link rel="stylesheet" type="text/css" media="%s" href="%s"/>\n' %
                  (media, xml_escape(cssfile)))
            w(u'<![endif]--> \n')
        # 4/ js files
        for jsfile in jsfiles:
            if skiphead:
                # Don't insert <script> tags directly as they would be
                # interpreted directly by some browsers (e.g. IE).
                # Use <cubicweb:script> tags instead and let
                # `loadAjaxHtmlHead` handle the script insertion / execution.
                w(u'<cubicweb:script src="%s"></cubicweb:script>\n' %
                  xml_escape(jsfile))
                # FIXME: a probably better implementation might be to add
                #        JS or CSS urls in a JS list that loadAjaxHtmlHead
                #        would iterate on and postprocess:
                #            cw._ajax_js_scripts.push('myscript.js')
                #        Then, in loadAjaxHtmlHead, do something like:
                #            jQuery.each(cw._ajax_js_script, jQuery.getScript)
            else:
                w(u'<script type="text/javascript" src="%s"></script>\n' %
                  xml_escape(jsfile))
        # 5/ post inlined scripts (i.e. scripts depending on other JS files)
        if self.post_inlined_scripts:
            if skiphead:
                for script in self.post_inlined_scripts:
                    w(u'<cubicweb:script>')
                    w(xml_escape(script))
                    w(u'</cubicweb:script>')
            else:
                w(self.script_opening)
                w(u'\n\n'.join(self.post_inlined_scripts))
                w(self.script_closing)
        # at the start of this function, the parent UStringIO may already have
        # data in it, so we can't w(u'<head>\n') at the top. Instead, we create
        # a temporary UStringIO to get the same debugging output formatting
        # if debugging is enabled.
        headtag = UStringIO(tracewrites=self.tracewrites)
        if not skiphead:
            headtag.write(u'<head>\n')
            w(u'</head>\n')
        return headtag.getvalue() + super(HTMLHead, self).getvalue()


class HTMLStream(object):
    """represents a HTML page.

    This is used my main templates so that HTML headers can be added
    at any time during the page generation.

    HTMLStream uses the (U)StringIO interface to be compliant with
    existing code.
    """

    def __init__(self, req):
        self.tracehtml = req.tracehtml
        # stream for <head>
        self.head = req.html_headers
        # main stream
        self.body = UStringIO(tracewrites=req.tracehtml)
        # this method will be assigned to self.w in views
        self.write = self.body.write
        self.doctype = u''
        self._htmlattrs = [('lang', req.lang)]
        # keep main_stream's reference on req for easier text/html demoting
        req.main_stream = self

    @deprecated('[3.17] there are no namespaces in html, xhtml is not served any longer')
    def add_namespace(self, prefix, uri):
        pass

    @deprecated('[3.17] there are no namespaces in html, xhtml is not served any longer')
    def set_namespaces(self, namespaces):
        pass

    def add_htmlattr(self, attrname, attrvalue):
        self._htmlattrs.append( (attrname, attrvalue) )

    def set_htmlattrs(self, attrs):
        self._htmlattrs = attrs

    def set_doctype(self, doctype, reset_xmldecl=None):
        self.doctype = doctype
        if reset_xmldecl is not None:
            warn('[3.17] xhtml is no more supported',
                 DeprecationWarning, stacklevel=2)

    @property
    def htmltag(self):
        attrs = ' '.join('%s="%s"' % (attr, xml_escape(value))
                         for attr, value in self._htmlattrs)
        if attrs:
            return '<html xmlns:cubicweb="http://www.cubicweb.org" %s>' % attrs
        return '<html xmlns:cubicweb="http://www.cubicweb.org">'

    def getvalue(self):
        """writes HTML headers, closes </head> tag and writes HTML body"""
        if self.tracehtml:
            css = u'\n'.join((u'span {',
                              u'  font-family: monospace;',
                              u'  word-break: break-all;',
                              u'  word-wrap: break-word;',
                              u'}',
                              u'span:hover {',
                              u'  color: red;',
                              u'  text-decoration: underline;',
                              u'}'))
            style = u'<style type="text/css">\n%s\n</style>\n' % css
            return (u'<!DOCTYPE html>\n'
                    + u'<html>\n<head>\n%s\n</head>\n' % style
                    + u'<body>\n'
                    + u'<span>' + xml_escape(self.doctype) + u'</span><br/>'
                    + u'<span>' + xml_escape(self.htmltag) + u'</span><br/>'
                    + self.head.getvalue()
                    + self.body.getvalue()
                    + u'<span>' + xml_escape(u'</html>') + u'</span>'
                    + u'</body>\n</html>')
        return u'%s\n%s\n%s\n%s\n</html>' % (self.doctype,
                                             self.htmltag,
                                             self.head.getvalue(),
                                             self.body.getvalue())


class CubicWebJsonEncoder(json.JSONEncoder):
    """define a json encoder to be able to encode yams std types"""

    def default(self, obj):
        if hasattr(obj, '__json_encode__'):
            return obj.__json_encode__()
        if isinstance(obj, datetime.datetime):
            return ustrftime(obj, '%Y/%m/%d %H:%M:%S')
        elif isinstance(obj, datetime.date):
            return ustrftime(obj, '%Y/%m/%d')
        elif isinstance(obj, datetime.time):
            return obj.strftime('%H:%M:%S')
        elif isinstance(obj, datetime.timedelta):
            return (obj.days * 24 * 60 * 60) + obj.seconds
        elif isinstance(obj, decimal.Decimal):
            return float(obj)
        elif isinstance(obj, Binary):
            return base64.b64encode(obj.getvalue()).decode('ascii')
        try:
            return json.JSONEncoder.default(self, obj)
        except TypeError:
            # we never ever want to fail because of an unknown type,
            # just return None in those cases.
            return None

def json_dumps(value, **kwargs):
    return json.dumps(value, cls=CubicWebJsonEncoder, **kwargs)


class JSString(str):
    """use this string sub class in values given to :func:`js_dumps` to
    insert raw javascript chain in some JSON string
    """

def _dict2js(d, predictable=False):
    if predictable:
        it = sorted(d.items())
    else:
        it = d.items()
    res = [js_dumps(key, predictable) + ': ' + js_dumps(val, predictable)
           for key, val in it]
    return '{%s}' % ', '.join(res)

def _list2js(l, predictable=False):
    return '[%s]' % ', '.join([js_dumps(val, predictable) for val in l])

def js_dumps(something, predictable=False):
    """similar as :func:`json_dumps`, except values which are instances of
    :class:`JSString` are expected to be valid javascript and will be output
    as is

    >>> js_dumps({'hop': JSString('$.hop'), 'bar': None}, predictable=True)
    '{bar: null, hop: $.hop}'
    >>> js_dumps({'hop': '$.hop'})
    '{hop: "$.hop"}'
    >>> js_dumps({'hip': {'hop': JSString('momo')}})
    '{hip: {hop: momo}}'
    """
    if isinstance(something, dict):
        return _dict2js(something, predictable)
    if isinstance(something, list):
        return _list2js(something, predictable)
    if isinstance(something, JSString):
        return something
    return json_dumps(something, sort_keys=predictable)

PERCENT_IN_URLQUOTE_RE = re.compile(r'%(?=[0-9a-fA-F]{2})')
def js_href(javascript_code):
    """Generate a "javascript: ..." string for an href attribute.

    Some % which may be interpreted in a href context will be escaped.

    In an href attribute, url-quotes-looking fragments are interpreted before
    being given to the javascript engine. Valid url quotes are in the form
    ``%xx`` with xx being a byte in hexadecimal form. This means that ``%toto``
    will be unaltered but ``%babar`` will be mangled because ``ba`` is the
    hexadecimal representation of 186.

    >>> js_href('alert("babar");')
    'javascript: alert("babar");'
    >>> js_href('alert("%babar");')
    'javascript: alert("%25babar");'
    >>> js_href('alert("%toto %babar");')
    'javascript: alert("%toto %25babar");'
    >>> js_href('alert("%1337%");')
    'javascript: alert("%251337%");'
    """
    return 'javascript: ' + PERCENT_IN_URLQUOTE_RE.sub(r'%25', javascript_code)


logger = getLogger('cubicweb.utils')

class QueryCache(object):
    """ a minimalist dict-like object to be used by the querier
    and native source (replaces lgc.cache for this very usage)

    To be efficient it must be properly used. The usage patterns are
    quite specific to its current clients.

    The ceiling value should be sufficiently high, else it will be
    ruthlessly inefficient (there will be warnings when this happens).
    A good (high enough) value can only be set on a per-application
    value. A default, reasonnably high value is provided but tuning
    e.g `rql-cache-size` can certainly help.

    There are two kinds of elements to put in this cache:
    * frequently used elements
    * occasional elements

    The former should finish in the _permanent structure after some
    warmup.

    Occasional elements can be buggy requests (server-side) or
    end-user (web-ui provided) requests. These have to be cleaned up
    when they fill the cache, without evicting the useful, frequently
    used entries.
    """
    # quite arbitrary, but we want to never
    # immortalize some use-a-little query
    _maxlevel = 15

    def __init__(self, ceiling=3000):
        self._max = ceiling
        # keys belonging forever to this cache
        self._permanent = set()
        # mapping of key (that can get wiped) to getitem count
        self._transient = {}
        self._data = {}
        self._lock = Lock()

    def __len__(self):
        with self._lock:
            return len(self._data)

    def __getitem__(self, k):
        with self._lock:
            if k in self._permanent:
                return self._data[k]
            v = self._transient.get(k, _MARKER)
            if v is _MARKER:
                self._transient[k] = 1
                return self._data[k]
            if v > self._maxlevel:
                self._permanent.add(k)
                self._transient.pop(k, None)
            else:
                self._transient[k] += 1
            return self._data[k]

    def __setitem__(self, k, v):
        with self._lock:
            if len(self._data) >= self._max:
                self._try_to_make_room()
            self._data[k] = v

    def pop(self, key, default=_MARKER):
        with self._lock:
            try:
                if default is _MARKER:
                    return self._data.pop(key)
                return self._data.pop(key, default)
            finally:
                if key in self._permanent:
                    self._permanent.remove(key)
                else:
                    self._transient.pop(key, None)

    def clear(self):
        with self._lock:
            self._clear()

    def _clear(self):
        self._permanent = set()
        self._transient = {}
        self._data = {}

    def _try_to_make_room(self):
        current_size = len(self._data)
        items = sorted(self._transient.items(), key=itemgetter(1))
        level = 0
        for k, v in items:
            self._data.pop(k, None)
            self._transient.pop(k, None)
            if v > level:
                datalen = len(self._data)
                if datalen == 0:
                    return
                if (current_size - datalen) / datalen > .1:
                    break
                level = v
        else:
            # we removed cruft but everything is permanent
            if len(self._data) >= self._max:
                logger.warning('Cache %s is full.' % id(self))
                self._clear()

    def _usage_report(self):
        with self._lock:
            return {'itemcount': len(self._data),
                    'transientcount': len(self._transient),
                    'permanentcount': len(self._permanent)}

    def popitem(self):
        raise NotImplementedError()

    def setdefault(self, key, default=None):
        raise NotImplementedError()

    def update(self, other):
        raise NotImplementedError()



# copyright 2003-2011 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""base xml and rss views"""


from cubicweb import _

from six.moves import range

from yams import xy

from cubicweb.schema import VIRTUAL_RTYPES
from cubicweb.view import EntityView
from cubicweb.web.views.xmlrss import SERIALIZERS

try:
    import rdflib
except ImportError:
    rdflib = None

if rdflib is not None:
    RDF = rdflib.Namespace('http://www.w3.org/1999/02/22-rdf-syntax-ns#')
    CW = rdflib.Namespace('http://ns.cubicweb.org/cubicweb/0.0/')
    from rdflib import Literal, URIRef, Namespace

    def urijoin(item):
        base, ext = item
        return URIRef(Namespace(base)[ext])

    SKIP_RTYPES = VIRTUAL_RTYPES | set(['cwuri', 'is', 'is_instance_of'])

    class RDFView(EntityView):
        """rdf view for entities"""
        __regid__ = 'rdf'
        title = _('rdf export')
        templatable = False
        binary = True
        format = 'xml'
        content_type = 'text/xml' # +rdf

        def call(self):
            graph = rdflib.Graph()
            graph.bind('cw', CW)
            for prefix, xmlns in xy.XY.prefixes.items():
                graph.bind(prefix, rdflib.Namespace(xmlns))
            for i in range(self.cw_rset.rowcount):
                entity = self.cw_rset.complete_entity(i, 0)
                self.entity2graph(graph, entity)
            self.w(graph.serialize(format=self.format))

        def entity_call(self, entity):
            self.call()

        def entity2graph(self, graph, entity):
            cwuri = URIRef(entity.cwuri)
            add = graph.add
            add( (cwuri, RDF.type, CW[entity.e_schema.type]) )
            try:
                for item in xy.xeq(entity.e_schema.type):
                    add( (cwuri, RDF.type, urijoin(item)) )
            except xy.UnsupportedVocabulary:
                pass
            for rschema, eschemas, role in entity.e_schema.relation_definitions('relation'):
                rtype = rschema.type
                if rtype in SKIP_RTYPES or rtype.endswith('_permission'):
                    continue
                for eschema in eschemas:
                    if eschema.final:
                        try:
                            value = entity.cw_attr_cache[rtype]
                        except KeyError:
                            continue # assuming rtype is Bytes
                        if value is not None:
                            add( (cwuri, CW[rtype], Literal(value)) )
                            try:
                                for item in xy.xeq('%s %s' % (entity.e_schema.type, rtype)):
                                    add( (cwuri, urijoin(item[1]), Literal(value)) )
                            except xy.UnsupportedVocabulary:
                                pass
                    else:
                        for related in entity.related(rtype, role, entities=True, safe=True):
                            if role == 'subject':
                                add( (cwuri, CW[rtype], URIRef(related.cwuri)) )
                                try:
                                    for item in xy.xeq('%s %s' % (entity.e_schema.type, rtype)):
                                        add( (cwuri, urijoin(item[1]), URIRef(related.cwuri)) )
                                except xy.UnsupportedVocabulary:
                                    pass
                            else:
                                add( (URIRef(related.cwuri), CW[rtype], cwuri) )


    class RDFN3View(RDFView):
        __regid__ = 'n3rdf'
        format = 'n3'
        content_type = 'text/n3'



# copyright 2003-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""unittests for cw.devtools.testlib module"""

from io import BytesIO, StringIO
from unittest import TextTestRunner

from six import PY2

from logilab.common.testlib import TestSuite, TestCase, unittest_main
from logilab.common.registry import yes

from cubicweb.devtools import htmlparser
from cubicweb.devtools.testlib import CubicWebTC
from cubicweb.pytestconf import clean_repo_test_cls

class FakeFormTC(TestCase):
    def test_fake_form(self):
        class entity:
            cw_etype = 'Entity'
            eid = 0
        sio = BytesIO(b'hop\n')
        form = CubicWebTC.fake_form('import',
                                    {'file': ('filename.txt', sio),
                                     'encoding': u'utf-8',
                                    }, [(entity, {'field': 'value'})])
        self.assertEqual(form, {'__form_id': 'import',
                                '__maineid': 0,
                                '__type:0': 'Entity',
                                '_cw_entity_fields:0': '__type,field',
                                '_cw_fields': 'encoding,file',
                                'eid': [0],
                                'encoding': u'utf-8',
                                'field:0': 'value',
                                'file': ('filename.txt', sio)})

class WebTestTC(TestCase):

    def setUp(self):
        output = BytesIO() if PY2 else StringIO()
        self.runner = TextTestRunner(stream=output)

    def test_error_raised(self):
        class MyWebTest(CubicWebTC):

            def test_error_view(self):
                with self.admin_access.web_request() as req:
                    req.create_entity('Bug', title=u"bt")
                    self.view('raising', req.execute('Bug B'), template=None, req=req)

            def test_correct_view(self):
                with self.admin_access.web_request() as req:
                    self.view('primary', req.execute('CWUser U'), template=None, req=req)

        tests = [MyWebTest('test_error_view'), MyWebTest('test_correct_view')]
        result = self.runner.run(TestSuite(tests))
        self.assertEqual(result.testsRun, 2)
        self.assertEqual(len(result.errors), 0)
        self.assertEqual(len(result.failures), 1)
        clean_repo_test_cls(MyWebTest)


class RepoInstancesConsistencyTC(CubicWebTC):
    test_db_id = 'RepoInstancesConsistencyTC'

    def pre_setup_database(self, cnx, config):
        self.assertIs(cnx.repo, config.repository())

    def test_pre_setup(self):
        pass


HTML_PAGE = u"""<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html>
  <head><title>need a title</title></head>
  <body>
    <h1>Hello World !</h1>
  </body>
</html>
"""

HTML_PAGE2 = u"""<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html>
 <head><title>need a title</title></head>
 <body>
   <h1>Test</h1>
   <h1>Hello <a href="http://www.google.com">world</a> !</h1>
   <h2>h2 title</h2>
   <h3>h3 title</h3>
   <h2>antoher h2 title</h2>
   <h4>h4 title</h4>
   <p><a href="http://www.logilab.org">Logilab</a> introduces CW !</p>
 </body>
</html>
"""

HTML_PAGE_ERROR = u"""<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html>
 <head><title>need a title</title></head>
 <body>
   Logilab</a> introduces CW !
 </body>
</html>
"""

HTML_NON_STRICT = u"""<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html>
  <head><title>need a title</title></head>
  <body>
    <h1><a href="something.com">title</h1>
  </body>
</html>
"""


class HTMLPageInfoTC(TestCase):
    """test cases for PageInfo"""

    def setUp(self):
        parser = htmlparser.HTMLValidator()
        # disable cleanup that would remove doctype
        parser.preprocess_data = lambda data: data
        self.page_info = parser.parse_string(HTML_PAGE2)

    def test_source1(self):
        """make sure source is stored correctly"""
        self.assertEqual(self.page_info.source, HTML_PAGE2)

    def test_source2(self):
        """make sure source is stored correctly - raise exception"""
        parser = htmlparser.DTDValidator()
        self.assertRaises(AssertionError, parser.parse_string, HTML_PAGE_ERROR)

    def test_has_title_no_level(self):
        """tests h? tags information"""
        self.assertEqual(self.page_info.has_title('Test'), True)
        self.assertEqual(self.page_info.has_title('Test '), False)
        self.assertEqual(self.page_info.has_title('Tes'), False)
        self.assertEqual(self.page_info.has_title('Hello world !'), True)

    def test_has_title_level(self):
        """tests h? tags information"""
        self.assertEqual(self.page_info.has_title('Test', level = 1), True)
        self.assertEqual(self.page_info.has_title('Test', level = 2), False)
        self.assertEqual(self.page_info.has_title('Test', level = 3), False)
        self.assertEqual(self.page_info.has_title('Test', level = 4), False)
        self.assertRaises(IndexError, self.page_info.has_title, 'Test', level = 5)

    def test_has_title_regexp_no_level(self):
        """tests has_title_regexp() with no particular level specified"""
        self.assertEqual(self.page_info.has_title_regexp('h[23] title'), True)

    def test_has_title_regexp_level(self):
        """tests has_title_regexp() with a particular level specified"""
        self.assertEqual(self.page_info.has_title_regexp('h[23] title', 2), True)
        self.assertEqual(self.page_info.has_title_regexp('h[23] title', 3), True)
        self.assertEqual(self.page_info.has_title_regexp('h[23] title', 4), False)

    def test_appears(self):
        """tests PageInfo.appears()"""
        self.assertEqual(self.page_info.appears('CW'), True)
        self.assertEqual(self.page_info.appears('Logilab'), True)
        self.assertEqual(self.page_info.appears('Logilab introduces'), True)
        self.assertEqual(self.page_info.appears('H2 title'), False)

    def test_has_link(self):
        """tests has_link()"""
        self.assertEqual(self.page_info.has_link('Logilab'), True)
        self.assertEqual(self.page_info.has_link('logilab'), False)
        self.assertEqual(self.page_info.has_link('Logilab', 'http://www.logilab.org'), True)
        self.assertEqual(self.page_info.has_link('Logilab', 'http://www.google.com'), False)

    def test_has_link_regexp(self):
        """test has_link_regexp()"""
        self.assertEqual(self.page_info.has_link_regexp('L[oi]gilab'), True)
        self.assertEqual(self.page_info.has_link_regexp('L[ai]gilab'), False)


class CWUtilitiesTC(CubicWebTC):

    def test_temporary_permissions_eschema(self):
        eschema = self.schema['CWUser']
        with self.temporary_permissions(CWUser={'read': ()}):
            self.assertEqual(eschema.permissions['read'], ())
            self.assertTrue(eschema.permissions['add'])
        self.assertTrue(eschema.permissions['read'], ())

    def test_temporary_permissions_rdef(self):
        rdef = self.schema['CWUser'].rdef('in_group')
        with self.temporary_permissions((rdef, {'read': ()})):
            self.assertEqual(rdef.permissions['read'], ())
            self.assertTrue(rdef.permissions['add'])
        self.assertTrue(rdef.permissions['read'], ())

    def test_temporary_permissions_rdef_with_exception(self):
        rdef = self.schema['CWUser'].rdef('in_group')
        try:
            with self.temporary_permissions((rdef, {'read': ()})):
                self.assertEqual(rdef.permissions['read'], ())
                self.assertTrue(rdef.permissions['add'])
                raise ValueError('goto')
        except ValueError:
            self.assertTrue(rdef.permissions['read'], ())
        else:
            self.fail('exception was caught unexpectedly')

    def test_temporary_appobjects_registered(self):

        class AnAppobject(object):
            __registries__ = ('hip',)
            __regid__ = 'hop'
            __select__ = yes()
            registered = None

            @classmethod
            def __registered__(cls, reg):
                cls.registered = reg

        with self.temporary_appobjects(AnAppobject):
            self.assertEqual(self.vreg['hip'], AnAppobject.registered)
            self.assertIn(AnAppobject, self.vreg['hip']['hop'])
        self.assertNotIn(AnAppobject, self.vreg['hip']['hop'])

    def test_login(self):
        """Calling login should not break hook control"""
        with self.admin_access.repo_cnx() as cnx:
            self.hook_executed = False
            self.create_user(cnx, 'babar')
            cnx.commit()

        from cubicweb.server import hook
        from cubicweb.predicates import is_instance

        class MyHook(hook.Hook):
            __regid__ = 'whatever'
            __select__ = hook.Hook.__select__ & is_instance('CWProperty')
            category = 'test-hook'
            events = ('after_add_entity',)
            test = self

            def __call__(self):
                self.test.hook_executed = True

        with self.new_access('babar').repo_cnx() as cnx:
            with self.temporary_appobjects(MyHook):
                with cnx.allow_all_hooks_but('test-hook'):
                    prop = cnx.create_entity('CWProperty', pkey=u'ui.language', value=u'en')
                    cnx.commit()
                    self.assertFalse(self.hook_executed)


class RepoAccessTC(CubicWebTC):

    def test_repo_connection(self):
        acc = self.new_access('admin')
        with acc.repo_cnx() as cnx:
            rset = cnx.execute('Any X WHERE X is CWUser')
            self.assertTrue(rset)

    def test_client_connection(self):
        acc = self.new_access('admin')
        with acc.client_cnx() as cnx:
            rset = cnx.execute('Any X WHERE X is CWUser')
            self.assertTrue(rset)

    def test_web_request(self):
        acc = self.new_access('admin')
        with acc.web_request(elephant='babar') as req:
            rset = req.execute('Any X WHERE X is CWUser')
            self.assertTrue(rset)
            self.assertEqual('babar', req.form['elephant'])

    def test_admin_access(self):
        with self.admin_access.client_cnx() as cnx:
            self.assertEqual('admin', cnx.user.login)


if __name__ == '__main__':
    unittest_main()



if confirm('use Europe/Paris as timezone?'):
    timezone = 'Europe/Paris'
else:
    import pytz
    while True:
        timezone = raw_input('enter your timezone')
        if timezone in pytz.common_timezones:
            break

dbdriver = repo.system_source.dbdriver
if dbdriver == 'postgres':
    sql("SET TIME ZONE '%s'" % timezone)

for entity in schema.entities():
    if entity.final or entity.type not in fsschema:
        continue
    change_attribute_type(entity.type, 'creation_date', 'TZDatetime', ask_confirm=False)
    change_attribute_type(entity.type, 'modification_date', 'TZDatetime', ask_confirm=False)

if dbdriver == 'postgres':
    sql("SET TIME ZONE UTC")



# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
import unittest

from cubicweb import devtools
from cubicweb.xy import xy

SKIPCAUSE = None
try:
    from cubicweb.spa2rql import Sparql2rqlTranslator
except ImportError as exc:
    SKIPCAUSE = str(exc)

xy.add_equivalence('Project', 'doap:Project')
xy.add_equivalence('Project creation_date', 'doap:Project doap:created')
xy.add_equivalence('Project name', 'doap:Project doap:name')
xy.add_equivalence('Project name', 'doap:Project dc:title')


@unittest.skipIf(SKIPCAUSE, SKIPCAUSE)
class XYTC(unittest.TestCase):

    schema = None

    @classmethod
    def setUpClass(cls):
        config = devtools.TestServerConfiguration('data', __file__)
        config.bootstrap_cubes()
        cls.schema = config.load_schema()

    def setUp(self):
        self.tr = Sparql2rqlTranslator(self.schema)

    def _test(self, sparql, rql, args={}):
        qi = self.tr.translate(sparql)
        self.assertEqual(qi.finalize(), (rql, args))

    def XXX_test_base_01(self):
        self._test('SELECT * WHERE { }', 'Any X')


    def test_base_is(self):
        self._test('''
    PREFIX doap: <http://usefulinc.com/ns/doap#>
    SELECT ?project
    WHERE  {
      ?project a doap:Project;
    }''', 'Any PROJECT WHERE PROJECT is Project')

    def test_base_rdftype(self):
        self._test('''
    PREFIX doap: <http://usefulinc.com/ns/doap#>
    PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
    SELECT ?project
    WHERE  {
      ?project rdf:type doap:Project.
    }''', 'Any PROJECT WHERE PROJECT is Project')

    def test_base_attr_sel(self):
        self._test('''
    PREFIX doap: <http://usefulinc.com/ns/doap#>
    SELECT ?created
    WHERE  {
      ?project a doap:Project;
              doap:created ?created.
    }''', 'Any CREATED WHERE PROJECT creation_date CREATED, PROJECT is Project')


    def test_base_attr_sel_distinct(self):
        self._test('''
    PREFIX doap: <http://usefulinc.com/ns/doap#>
    SELECT DISTINCT ?name
    WHERE  {
      ?project a doap:Project;
              doap:name ?name.
    }''', 'DISTINCT Any NAME WHERE PROJECT name NAME, PROJECT is Project')


    def test_base_attr_sel_reduced(self):
        self._test('''
    PREFIX doap: <http://usefulinc.com/ns/doap#>
    SELECT REDUCED ?name
    WHERE  {
      ?project a doap:Project;
              doap:name ?name.
    }''', 'Any NAME WHERE PROJECT name NAME, PROJECT is Project')


    def test_base_attr_sel_limit_offset(self):
        self._test('''
    PREFIX doap: <http://usefulinc.com/ns/doap#>
    SELECT ?name
    WHERE  {
      ?project a doap:Project;
              doap:name ?name.
    }
    LIMIT 20''', 'Any NAME LIMIT 20 WHERE PROJECT name NAME, PROJECT is Project')
        self._test('''
    PREFIX doap: <http://usefulinc.com/ns/doap#>
    SELECT ?name
    WHERE  {
      ?project a doap:Project;
              doap:name ?name.
    }
    LIMIT 20 OFFSET 10''', 'Any NAME LIMIT 20 OFFSET 10 WHERE PROJECT name NAME, PROJECT is Project')


    def test_base_attr_sel_orderby(self):
        self._test('''
    PREFIX doap: <http://usefulinc.com/ns/doap#>
    SELECT ?name
    WHERE  {
      ?project a doap:Project;
              doap:name ?name;
              doap:created ?created.
    }
    ORDER BY ?name DESC(?created)''', 'Any NAME ORDERBY NAME ASC, CREATED DESC WHERE PROJECT name NAME, PROJECT creation_date CREATED, PROJECT is Project')


    def test_base_any_attr_sel(self):
        self._test('''
    PREFIX dc: <http://purl.org/dc/elements/1.1/>
    SELECT ?x ?cd
    WHERE  {
      ?x dc:date ?cd;
    }''', 'Any X, CD WHERE X creation_date CD')


    def test_base_any_attr_sel_amb(self):
        xy.add_equivalence('Version publication_date', 'doap:Version dc:date')
        try:
            self._test('''
    PREFIX dc: <http://purl.org/dc/elements/1.1/>
    SELECT ?x ?cd
    WHERE  {
      ?x dc:date ?cd;
    }''', '(Any X, CD WHERE , X creation_date CD) UNION (Any X, CD WHERE , X publication_date CD, X is Version)')
        finally:
            xy.remove_equivalence('Version publication_date', 'doap:Version dc:date')


    def test_base_any_attr_sel_amb_limit_offset(self):
        xy.add_equivalence('Version publication_date', 'doap:Version dc:date')
        try:
            self._test('''
    PREFIX dc: <http://purl.org/dc/elements/1.1/>
    SELECT ?x ?cd
    WHERE  {
      ?x dc:date ?cd;
    }
    LIMIT 20 OFFSET 10''', 'Any X, CD LIMIT 20 OFFSET 10 WITH X, CD BEING ((Any X, CD WHERE , X creation_date CD) UNION (Any X, CD WHERE , X publication_date CD, X is Version))')
        finally:
            xy.remove_equivalence('Version publication_date', 'doap:Version dc:date')


    def test_base_any_attr_sel_amb_orderby(self):
        xy.add_equivalence('Version publication_date', 'doap:Version dc:date')
        try:
            self._test('''
    PREFIX dc: <http://purl.org/dc/elements/1.1/>
    SELECT ?x ?cd
    WHERE  {
      ?x dc:date ?cd;
    }
    ORDER BY DESC(?cd)''', 'Any X, CD ORDERBY CD DESC WITH X, CD BEING ((Any X, CD WHERE , X creation_date CD) UNION (Any X, CD WHERE , X publication_date CD, X is Version))')
        finally:
            xy.remove_equivalence('Version publication_date', 'doap:Version dc:date')


    def test_restr_attr(self):
        self._test('''
    PREFIX doap: <http://usefulinc.com/ns/doap#>
    SELECT ?project
    WHERE  {
      ?project a doap:Project;
              doap:name "cubicweb".
    }''', 'Any PROJECT WHERE PROJECT name %(a)s, PROJECT is Project', {'a': 'cubicweb'})

    def test_dctitle_both_project_cwuser(self):
        self._test('''
    PREFIX doap: <http://usefulinc.com/ns/doap#>
    PREFIX dc: <http://purl.org/dc/elements/1.1/>
    SELECT ?project ?title
    WHERE  {
      ?project a doap:Project;
              dc:title ?title.
    }''', 'Any PROJECT, TITLE WHERE PROJECT name TITLE, PROJECT is Project')

# # Two elements in the group
# PREFIX :  <http://example.org/ns#>
# SELECT *
# { :p :q :r  OPTIONAL { :a :b :c }
#   :p :q :r  OPTIONAL { :a :b :c }
# }

# PREFIX : <http://example.org/ns#>
# SELECT *
# {
#   { ?s ?p ?o } UNION { ?a ?b ?c }
# }

# PREFIX dob: <http://placetime.com/interval/gregorian/1977-01-18T04:00:00Z/P>
# PREFIX time: <http://www.ai.sri.com/daml/ontologies/time/Time.daml#>
# PREFIX dc: <http://purl.org/dc/elements/1.1/>
# SELECT ?desc
# WHERE  {
#   dob:1D a time:ProperInterval;
#          dc:description ?desc.
# }

if __name__ == '__main__':
    unittest.main()



# copyright 2003-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
from yams.buildobjs import (EntityType, RelationDefinition, String, SubjectRelation,
                            ComputedRelation, Int)
from cubicweb.schema import ERQLExpression


class Person(EntityType):
    name = String()


class Affaire(EntityType):
    __permissions__ = {
        'read':   ('managers',
                   ERQLExpression('X owned_by U'), ERQLExpression('X concerne S?, S owned_by U')),
        'add':    ('managers', ERQLExpression('X concerne S, S owned_by U')),
        'update': ('managers', 'owners', ERQLExpression('X in_state S, S name in ("pitetre", "en cours")')),
        'delete': ('managers', 'owners', ERQLExpression('X concerne S, S owned_by U')),
        }
    ref = String(fulltextindexed=True, indexed=True, maxsize=16)
    documented_by = SubjectRelation('Card', cardinality='1*')
    concerne = SubjectRelation(('Societe', 'Note'), cardinality='1*')


class Societe(EntityType):
    __permissions__ = {
        'read': ('managers', 'users', 'guests'),
        'update': ('managers', 'owners', ERQLExpression('U login L, X nom L')),
        'delete': ('managers', 'owners', ERQLExpression('U login L, X nom L')),
        'add': ('managers', 'users',)
        }
    nom = String()


class Division(Societe):
    __specializes_schema__ = True


class Note(EntityType):
    pass


class require_permission(RelationDefinition):
    subject = ('Card', 'Note')
    object = 'CWPermission'


class require_state(RelationDefinition):
    subject = 'CWPermission'
    object = 'State'


class inlined_card(RelationDefinition):
    subject = 'Affaire'
    object = 'Card'
    inlined = True
    cardinality = '?*'

class inlined_note(RelationDefinition):
    subject = 'Card'
    object = 'Note'
    inlined = True
    cardinality = '?*'

class inlined_affaire(RelationDefinition):
    subject = 'Note'
    object = 'Affaire'
    inlined = True
    cardinality = '?*'

class responsable(RelationDefinition):
    subject = 'Societe'
    object = 'CWUser'
    inlined = True
    cardinality = '1*'

class Contribution(EntityType):
    code = Int()

class ArtWork(EntityType):
    name = String()

class Role(EntityType):
    name = String()

class contributor(RelationDefinition):
    subject = 'Contribution'
    object = 'Person'
    cardinality = '1*'
    inlined = True

class manifestation(RelationDefinition):
    subject = 'Contribution'
    object = 'ArtWork'

class role(RelationDefinition):
    subject = 'Contribution'
    object = 'Role'

class illustrator_of(ComputedRelation):
    rule = ('C is Contribution, C contributor S, C manifestation O, '
            'C role R, R name "illustrator"')

class participated_in(ComputedRelation):
    rule = 'S contributor O'

class match(RelationDefinition):
    subject = 'ArtWork'
    object = 'Note'
